#
#  ktool | ktool
#  macho.py
#
#  This file contains utilities for basic parsing of MachO File headers and such.
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#

import os
from collections import namedtuple
from enum import Enum
from io import BytesIO
from typing import Tuple, Dict, Union, BinaryIO

from kmacho import *
from kmacho.structs import *
from ktool.exceptions import *
from ktool.util import log
mmap = None


class MachOFileType(Enum):
    FAT = 0
    THIN = 1


class BackingFile:
    def __init__(self, fp: Union[BinaryIO, BytesIO], use_mmaped_io=True):
        self.fp = fp

        if hasattr(fp, 'name'):
            self.name = os.path.basename(fp.name)
        else:
            self.name = ''
        if use_mmaped_io:
            # noinspection PyBroadException
            try:
                assert not isinstance(fp, BytesIO)
                global mmap
                import mmap
                self.file = mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_COPY)
            except Exception:
                use_mmaped_io = False
            f = fp
            old_file_position = f.tell()
            f.seek(0, os.SEEK_END)
            self.size = f.tell()
            f.seek(old_file_position)

        if not use_mmaped_io:
            self.file = bytearray(fp.read())
            self.size = len(self.file)

    def read_bytes(self, location, count):
        return bytes(self.file[location:location+count])

    def read_int(self, location, count):
        return int.from_bytes(self.read_bytes(location, count), "big")

    def write(self, location, data: bytes):
        data = bytearray(data)

        if isinstance(self.file, bytearray):
            count = len(data)
            for i in range(count):
                self.file[location+i] = data[i]

        else:
            # noinspection PyUnresolvedReferences
            assert isinstance(self.file, mmap.mmap)
            self.file.seek(location)
            self.file.write(data)
            self.file.seek(0)

    def close(self):
        self.fp.close()


class SlicedBackingFile:
    def __init__(self, backing_file: BackingFile, offset, size):
        self.file = bytearray(backing_file.read_bytes(offset, size))
        self.size = size
        self.name = backing_file.name

    def read_bytes(self, location, count):
        return bytes(self.file[location:location+count])

    def read_int(self, location, count):
        return int.from_bytes(self.read_bytes(location, count), "big")

    def write(self, location, data: bytes):
        count = len(data)
        for i in range(count):
            self.file[location + i] = data[i]


class MachOFile:
    def __init__(self, file, use_mmaped_io=True):
        self.file_object = file

        self.uses_mmaped_io = use_mmaped_io

        if hasattr(file, 'name'):
            self.filename = os.path.basename(file.name)
        else:
            self.filename = ''

        self.file = BackingFile(file, use_mmaped_io)

        self.slices = []

        self.magic = self.file.read_int(0, 4)

        if self.magic in [FAT_MAGIC, FAT_CIGAM]:
            self.type = MachOFileType.FAT
        elif self.magic in [MH_MAGIC, MH_CIGAM, MH_MAGIC_64, MH_CIGAM_64]:
            self.type = MachOFileType.THIN
        else:
            log.debug(f'Bad Magic: {hex(self.magic)}')
            raise UnsupportedFiletypeException

        if self.type == MachOFileType.FAT:
            self.header: fat_header = self._load_struct(0, fat_header, "big")
            for off in range(0, self.header.nfat_archs):
                offset = fat_header.SIZE + (off * fat_arch.SIZE)
                arch_struct: fat_arch = self._load_struct(offset, fat_arch, "big")
                sliced_backing_file = SlicedBackingFile(self.file, arch_struct.offset, arch_struct.size)
                log.debug_more(str(arch_struct))
                self.slices.append(Slice(self, sliced_backing_file, arch_struct))
        else:
            self.slices.append(Slice(self, self.file, None))

    def _load_struct(self, address: int, struct_type, endian="little"):
        size = struct_type.SIZE
        data = self.file.read_bytes(address, size)

        struct = Struct.create_with_bytes(struct_type, data, endian)
        struct.off = address

        return struct

    def __del__(self):
        self.file.close()


class Section:
    """

    """

    def __init__(self, segment, cmd):
        self.cmd = cmd
        self.segment = segment
        self.name = cmd.sectname
        self.vm_address = cmd.addr
        self.file_address = cmd.offset
        self.size = cmd.size

    def serialize(self):
        return {
            'command': self.cmd.serialize(),
            'name': self.name,
            'vm_address': self.vm_address,
            'file_address': self.file_address,
            'size': self.size
        }


class Segment:
    """

    """

    def __init__(self, image, cmd):
        self.image = image
        self.is64 = isinstance(cmd, segment_command_64)
        self.cmd = cmd
        self.vm_address = cmd.vmaddr
        self.file_address = cmd.fileoff
        self.size = cmd.vmsize
        self.name = cmd.segname

        self.sections: Dict[str, Section] = self._process_sections()

        self.type = SectionType(S_FLAGS_MASKS.SECTION_TYPE & self.cmd.flags)

    def __str__(self):
        return f'Segment {self.name} at {hex(self.vm_address)}\n'

    def serialize(self):
        segment = {
            'command': self.cmd.serialize(),
            'name': self.name,
            'vm_address': self.vm_address,
            'file_address': self.file_address,
            'size': self.size,
            'type': self.type.name,
        }
        sects = {}
        for section_name, sect in self.sections.items():
            sects[section_name] = sect.serialize()
        segment['sections'] = sects
        return segment

    def _process_sections(self) -> Dict[str, Section]:
        sections = {}
        ea = self.cmd.off + self.cmd.SIZE

        for sect in range(0, self.cmd.nsects):
            sect = self.image.load_struct(ea, section_64 if self.is64 else section)
            sect = Section(self, sect)
            sections[sect.name] = sect
            ea += section_64.SIZE if self.is64 else section.SIZE

        return sections


class VM:
    """
    New Virtual Address translation based on actual VM -> physical pages

    """

    def __init__(self, page_size):
        self.page_size = page_size
        self.page_size_bits = (self.page_size - 1).bit_length()
        self.page_table = {}
        self.tlb = {}
        self.vm_base_addr = None
        self.dirty = False

        self.fallback: Union[None, MisalignedVM] = None

        self.detag_kern_64 = False
        self.detag_64 = False

    def vm_check(self, address):
        try:
            self.translate(address)
            return True
        except ValueError:
            return False

    def add_segment(self, segment: Segment):
        if segment.name == '__PAGEZERO':
            return

        self.fallback.add_segment(segment)

        if self.vm_base_addr is None:
            self.vm_base_addr = segment.vm_address
        if self.dirty or segment.vm_address % self.page_size != 0:
            self.dirty = True
            log.warn(f'MachO File could not be aligned.')
            raise MachOAlignmentError
        self.map_pages(segment.file_address, segment.vm_address, segment.size)

    def translate(self, address) -> int:

        l_addr = address

        if self.detag_kern_64:
            address = address | (0xFFFF << 12*4)

        if self.detag_64:
            address = address & 0xFFFFFFFFF

        try:
            return self.tlb[address]
        except KeyError:
            pass

        page_offset = address & self.page_size - 1
        page_location = address >> self.page_size_bits
        try:
            phys_page = self.page_table[page_location]
            physical_location = phys_page + page_offset
            self.tlb[address] = physical_location
            return physical_location
        except KeyError:

            log.info(f'Address {hex(address)} not mapped, attempting fallback')

            try:
                return self.fallback.translate(address)
            except ValueError:
                raise ValueError(f'Address {hex(address)} ({hex(l_addr)}) not in VA Table or fallback map. (page: {hex(page_location)})')

    def map_pages(self, physical_addr, virtual_addr, size):
        for i in range(size // self.page_size):
            self.page_table[virtual_addr + (i * self.page_size) >> self.page_size_bits] = physical_addr + (
                        i * self.page_size)


vm_obj = namedtuple("vm_obj", ["vmaddr", "vmend", "size", "fileaddr", "name"])


class MisalignedVM:
    """
    This is the manual backup if the image cant be mapped to 16/4k segments
    """

    def __init__(self, macho_slice):
        self.slice = macho_slice

        self.detag_kern_64 = False
        self.detag_64 = False

        self.map = {}
        self.stats = {}
        self.vm_base_addr = 0
        self.sorted_map = {}
        self.cache = {}

    def vm_check(self, vm_address):
        try:
            self.translate(vm_address)
            return True
        except ValueError:
            return False

    def translate(self, vm_address: int) -> int:

        if self.detag_kern_64:
            vm_address = vm_address | (0xFFFF << 12*4)

        if self.detag_64:
            vm_address = vm_address & 0xFFFFFFFFF

        if vm_address in self.cache:
            return self.cache[vm_address]

        for o in self.map.values():
            # noinspection PyChainedComparisons
            if vm_address >= o.vmaddr and o.vmend >= vm_address:
                file_addr = o.fileaddr + vm_address - o.vmaddr
                self.cache[vm_address] = file_addr
                return file_addr

        raise ValueError(f'Address {hex(vm_address)} couldn\'t be found in vm address set')

    def add_segment(self, segment: Segment):
        if segment.file_address == 0 and segment.size != 0:
            self.vm_base_addr = segment.vm_address

        if len(segment.sections) == 0:
            seg_obj = vm_obj(segment.vm_address, segment.vm_address + segment.size, segment.size, segment.file_address,
                             segment.name)
            log.info(str(seg_obj))
            self.map[segment.name] = seg_obj
            self.stats[segment.name] = 0
        else:
            for section in segment.sections.values():
                name = section.name if section.name not in self.map.keys() else section.name + '2'
                sect_obj = vm_obj(section.vm_address, section.vm_address + section.size, section.size,
                                  section.file_address, name)
                log.info(str(sect_obj))
                self.map[name] = sect_obj
                self.sorted_map = {k: v for k, v in sorted(self.map.items(), key=lambda item: item[1].vmaddr)}
                self.stats[name] = 0


class Slice:
    def __init__(self, macho_file, sliced_backing_file: Union[BackingFile, SlicedBackingFile], arch_struct: fat_arch=None, offset=0):
        self.file = sliced_backing_file
        self.macho_file = macho_file
        self.arch_struct: fat_arch = arch_struct

        if self.arch_struct:
            self.offset = arch_struct.offset
            self.type = self._load_type()
            self.subtype = self._load_subtype(self.type)
        else:
            self.offset = offset
            hdr = Struct.create_with_bytes(mach_header, self.get_bytes_at(0, 28))
            self.arch_struct = Struct.create_with_values(fat_arch, [hdr.cpu_type, hdr.cpu_subtype, 0, 0, 0])
            self.type = self._load_type()
            self.subtype = self._load_subtype(self.type)

        self.size = sliced_backing_file.size

        # noinspection PyArgumentList
        self.byte_order = "little" if self.get_int_at(0, 4, "little") in [MH_MAGIC, MH_MAGIC_64] else "big"

        self._cstring_cache = {}

    def patch(self, address: int, raw: bytes):
        log.debug_tm(f'Wrote {str(raw)} @ {address}')
        self.file.write(address, raw)
        assert self.file.read_bytes(address, len(raw)) == raw

    def full_bytes_for_slice(self):
        return bytes(self.file.read_bytes(0, self.file.size))

    def find(self, pattern: Union[str, bytes]):
        if isinstance(pattern, str):
            pattern = pattern.encode('utf-8')

        return self.file.file.find(pattern)

    def load_struct(self, addr: int, struct_type, endian="little"):
        size = struct_type.SIZE
        data = self.get_bytes_at(addr, size)

        struct = Struct.create_with_bytes(struct_type, data, endian)
        struct.off = addr

        return struct

    def get_int_at(self, addr: int, count: int, endian="little"):
        return int.from_bytes(self.file.read_bytes(addr, count), endian)

    def get_bytes_at(self, addr: int, count: int):
        return self.file.read_bytes(addr, count)

    def get_str_at(self, addr: int, count: int, force=False) -> str:
        if force:
            data = self.file.file[addr:addr + count]
            string = ""

            for ch in data:
                try:
                    string += bytes(ch).decode()
                except UnicodeDecodeError:
                    string += "?"
            return string

        return self.file.file[addr:addr + count].decode().rstrip('\x00')

    def get_cstr_at(self, addr: int, limit: int = 0):
        ea = addr

        if addr in self._cstring_cache:
            return self._cstring_cache[addr]

        count = 0
        while True:
            if self.file.file[ea] != 0:
                count += 1
                ea += 1
            else:
                break

        text = self.get_str_at(addr, count)

        self._cstring_cache[addr] = text

        return text

    def decode_uleb128(self, readHead: int) -> Tuple[int, int]:

        value = 0
        shift = 0

        while True:

            byte = self.get_int_at(readHead, 1)

            value |= (byte & 0x7f) << shift

            readHead += 1
            shift += 7

            if (byte & 0x80) == 0:
                break

        return value, readHead

    def _load_type(self) -> CPUType:
        cpu_type = self.arch_struct.cpu_type

        return CPUType(cpu_type)

    def _load_subtype(self, cputype: CPUType):
        cpu_subtype = self.arch_struct.cpu_subtype
        cpu_subtype = cpu_subtype & 0xFFFF

        try:
            sub = CPU_SUBTYPES[cputype]
            return sub(cpu_subtype)

        except KeyError:
            log.error(f'Unknown CPU SubType ({hex(cpu_subtype)}) ({self.arch_struct}). File an issue at '
                      f'https://github.com/cxnder/ktool')

            return CPUSubTypeARM64.ALL



#
#  ktool | ktool
#  window.py
#
#  This file houses Command Line GUI rendering code. A lot of it.
#  If you're looking to contribute to this right now, and are lost, please hit me up;
#       this file is a slight mess, and the entire UI framework is custom.
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#

# # # # #
#
# Comments:::
#   I'm not a huge fan of python-curses' cryptic, C-style abstractions, so I abstracted them out myself with a proper
#       OOP approach, which also serves to fix the curses (Y, X) coordinate handling crap.
#
# CURRENT TO-DO LIST:::
# TODO: Properly Abstract out Mouse Clicks / clean up mouse handler code
#
# CURRENT FEATURE TO-DO LIST:::
# TODO: Implement the title bar menu actions
#
# # # # #

import curses
import os
import pprint
from datetime import datetime
from math import ceil

from pygments import highlight
from pygments.formatters.terminal import TerminalFormatter
from pygments.formatters.terminal256 import Terminal256Formatter
from pygments.lexers.objective import ObjectiveCLexer

from kmacho import LOAD_COMMAND

from ktool.macho import MachOFile
from ktool.dyld import Dyld
from ktool.objc import ObjCImage
from ktool.headers import HeaderGenerator
from ktool.swift import load_swift_types, SwiftClass
from ktool.util import Table

from ktool.kcache import KernelCache, Kext

VERT_LINE = '│'
WINDOW_NAME = 'ktool'

BOX_CHARS = ['┦', '─', '━', '│', '┃', '┄', '┅', '┆', '┇', '┈', '┉', '┊', '┋', '┌', '┍', '┎', '┏', '┐', '┑', '┒', '┓', '└', '┕', '┖', '┗', '┘', '┙', '┚', '┛', '├', '┝', '┞', '┟', '┠', '┡', '┢', '┣', '┤', '┥', '┦', '┧', '┨', '┩', '┪', '┫', '┬', '┭', '┮', '┯', '┰', '┱', '┲', '┳', '┴', '┵', '┶', '┷', '┸', '┹', '┺', '┻', '┼', '┽', '┾', '┿', '╀', '╁', '╂', '╃', '╄', '╅', '╆', '╇', '╈', '╉', '╊', '╋', '╌', '╍', '╎', '╏', '═', '║', '╒', '╓', '╔', '╕', '╖', '╗', '╘', '╙', '╚', '╛', '╜', '╝', '╞', '╟', '╠', '╡', '╢', '╣', '╤', '╥', '╦', '╧', '╨', '╩', '╪', '╫', '╬', '╭', '╮', '╯', '╰', '╱', '╲', '╳', '╴', '╵', '╶', '╷', '╸', '╹', '╺', '╻', '╼', '╽', '╾', '╿']

SIDEBAR_WIDTH = 40

MAIN_TEXT = """ktool ------

This is a *very* pre-release version of the GUI Tool, and it has a long ways to go. 

Stay Updated with `python3 -m pip install --upgrade k2l` !

Mouse support is a WIP; quite a few things support mouse interaction already.

Navigate the sidebar with arrow keys or mouse. You can use left/right arrow, or spacebar, to expand/collapse submenus.

Hit tab to swap between the sidebar context and main context. Scroll the main context with up/down keys.

Backspace to exit (or click the X in the top right corner).
"""

PANIC_STRING = ""


def panic(msg):
    global PANIC_STRING
    PANIC_STRING = msg
    raise PanicException


ATTR_STRING_DEBUG = False


class ColorRep:
    def __init__(self, n):
        self.n = n

    def get_attr(self):
        return curses.color_pair(self.n)


# somewhat unused
class Attribute:
    HIGHLIGHTED = curses.A_STANDOUT
    UNDERLINED = curses.A_UNDERLINE
    COLOR_1 = ColorRep(1)
    COLOR_2 = ColorRep(2)
    COLOR_3 = ColorRep(3)
    COLOR_4 = ColorRep(4)
    COLOR_5 = ColorRep(5)
    COLOR_6 = ColorRep(6)
    COLOR_7 = ColorRep(7)


class AttributedString:
    def __init__(self, string: str):
        self.string = string
        self.attrs = []

    @staticmethod
    def ansi_to_attrstr(ansi_str):
        """This function translates ansi escaped strings (or manually specified ones, by replacing the "escape[" with §),
                to our Attributed String Format.

        :param ansi_str:
        :return:
        """
        pos = 0
        ansi_str = list(ansi_str)

        while pos < len(ansi_str):
            if ord(ansi_str[pos]) == 27:
                ansi_str[pos] = "§"
            pos += 1

        ansi_str = "".join(ansi_str)
        ansi_str = ansi_str.replace('§[', '§')

        if ATTR_STRING_DEBUG:
            return ansi_str

        pos = 0
        bland_pos = 0
        attr_str = AttributedString(ansi_str)
        bland_str = ""
        attr_start = 0
        attr_end = 0
        attr_color = 0
        while pos < len(ansi_str):
            if ansi_str[pos] == "§":
                ansi_escape_code = ""
                pos += 1
                if pos == len(ansi_str):
                    panic(ansi_str)
                while ansi_str[pos] != 'm':
                    ansi_escape_code += ansi_str[pos]
                    pos += 1
                ansi_list = ansi_escape_code.split(';')

                is_reset = False
                first_item = ansi_list[0]
                try:
                    int(first_item)
                except ValueError:
                    panic(str(ansi_list) + '\n' + ansi_str)
                if first_item == '38':
                    attr_color = AttributedString.fix_256_code(int(ansi_list[2]))
                elif first_item == '39' or first_item == '0':
                    is_reset = True
                elif 30 <= int(first_item) <= 37:
                    attr_color = int(first_item) - 30 + 8

                if is_reset:
                    attr_end = bland_pos
                    attr_str.set_attr(attr_start, attr_end, curses.color_pair(attr_color))
                else:
                    attr_start = bland_pos
                    attr_str.set_attr(attr_end, attr_start, curses.A_NORMAL)

                pos += 1
            else:
                bland_str += ansi_str[pos]
                bland_pos += 1
                pos += 1
        attr_str.string = bland_str
        return attr_str

    @staticmethod
    def fix_256_code(code):
        """Pygments 256 formatter sucks.

        :param code:
        :return:
        """
        if code == 125:
            return 168
        if code == 21:
            return 151
        if code == 28:
            return 118
        return code

    def set_attr(self, start, end, attr):
        self.attrs.append([[start, end], attr])

    def __str__(self):
        return self.string


# # # # #
#
# Custom Exceptions:::
# Almost all of these are handled internally.
#
# # # # #


class ExitProgramException(Exception):
    """Raise this within the run-loop to cleanly exit the program
    """

    def __init__(self):
        pass


class RebuildAllException(Exception):
    """Raise this to invoke a rebuild
    """


class PresentDebugMenuException(Exception):
    """Raise this within the runloop to present the debug menu
    """


class PresentTitleMenuException(Exception):
    """Raise this within the runloop to invoke the Title Bar Menu Rendering code
    """


class FileBrowserOpenNewFileException(Exception):
    """

    """


class HelpMenuException(Exception):
    """"""


class DestroyTitleMenuException(Exception):
    """Raise this to destroy the menu overlay
    """


class PanicException(Exception):
    """Raise this within the program and set the global PANIC_STRING to panic the window,
            and print the string after cleaning up the window display.
    """


class HexDumpTable(Table):
    """
    Subclass of table, just set the .hex value to a bytearray and it'll handle rendering it.

    """

    def __init__(self):
        super().__init__()
        self.titles = ['Raw Data', 'ASCII']
        self.hex = bytearray(b'')

    def fetch(self, row_start, row_count, screen_width):

        col_count = 2
        self.rows = []

        stack = ""
        decode_stack = ""
        stack_div = ""
        decode_stack_div = ""

        for i, byte in enumerate(self.hex[row_start*8:row_start*8+row_count*8]):
            stack_div += hex(byte)[2:].rjust(2, '0')
            decode_stack_div += byte.to_bytes(1, 'big').decode('ascii') + ' ' if byte in range(32, 127) else '. '
            if len(stack_div) >= 8:
                stack += stack_div + '  '
                decode_stack += decode_stack_div + '  '
                stack_div = ""
                decode_stack_div = ""
            if len(stack) >= 10 * col_count:
                self.rows.append([stack, decode_stack])
                stack = ""
                decode_stack = ""

        self.rows.append([stack, decode_stack])

        if not len(self.column_maxes) > 0:
            self.preheat()

        fetched = super().fetch(0, row_count, screen_width)

        self.rendered_row_cache = {}

        return fetched


class LazilyProcessedTextBuffer:
    def __init__(self):
        self.lines = []
        self.processed = False

        self.target = None
        self.target_args = []

    def go(self):
        self.lines = self.target(*self.target_args)
        self.processed = True


# # # # #
#
# Lower Level Display Abstraction
#
# # # # #


class RootBox:
    """
    The Root Box is an abstraction of the regular box, with no bounds/x-y coordinates

    It represents the entire terminal window itself and handles actually writing to/from the curses standard screen.
    """

    def __init__(self, stdscr):
        self.stdscr = stdscr

    def coord_translate(self, x, y):
        return x, y

    def write(self, x, y, string, attr):
        try:
            self.stdscr.addstr(y, x, string, attr)
        except curses.error:
            global PANIC_STRING
            PANIC_STRING = f'Rendering Error while writing {string} @ {x}, {y}\nScreen Bounds: {curses.COLS}x{curses.LINES}\nProgram Panicked'
            raise PanicException


class Box:
    """
    A Box is an abstraction of an area on the screen to write to.

    It's defined with a standard set of coords/dimensions, and writes to it are relative from those defined dimensions
    """

    def __init__(self, parent, x, y, width, height):
        self.parent = parent

        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def coord_translate(self, x, y):
        px, py = self.parent.coord_translate(self.x, self.y)
        return x - px, y - py

    def write(self, x, y, string, attr):
        self.parent.write(self.x + x, self.y + y, string, attr)

    def is_click_inbounds(self, x, y):
        min_x = self.x
        max_x = self.x + self.width

        min_y = self.y
        max_y = self.y + self.width

        if min_x <= x <= max_x:
            if min_y <= y <= max_y:
                return True

        return False


class ScrollingDisplayBuffer:
    """
    A ScrollingDisplayBuffer is a text-rendering abstraction to be placed within a standard Box.

    Set its `.lines` attribute, call .draw_lines(), and it will render those lines, cleanly wrapping them and
        implementing scrolling logic to move up and down the rendered buffer.
    """

    def __init__(self, parent, x, y, width, height):
        self.box = Box(parent, x, y + 1, width, height)
        self.scrollcursor = 0
        self.parent = parent
        self.x = parent.x
        self.y = parent.y
        self.width = width
        self.height = height

        self.render_attr = curses.A_NORMAL

        self.lines = []
        self.processed_lines = []
        self.pinned_lines = []

        self.wrap = True
        self.clean_wrap = True

        self.filled_line_count = 0

    @staticmethod
    def find_clean_breakpoint(text, maxwidth):
        """Find a clean place to wrap a line, if one exists

        :param text:
        :param maxwidth:
        :return:
        """
        if isinstance(text, AttributedString):
            text = text.string
        max_text = text[:maxwidth]
        max_text = max_text.strip()
        break_index = maxwidth
        for bindex, c in enumerate(max_text[::-1]):
            if c == ' ':
                break_index = maxwidth - bindex
                break
        if break_index == maxwidth:
            # we didn't find a space breakpoint, look for a / one
            for bindex, c in enumerate(max_text[::-1]):
                if c == '/':
                    break_index = maxwidth - bindex - 1
                    break
        return break_index

    def process_lines(self):
        """Process raw lines into a cleanly wrapped version that fits within our buffer's bounds.

        This method is intensive, *do not!* call it on every redraw, it will destroy performance. Call it as little
            as possible

        :return:
        """
        self.pinned_lines = []
        if self.wrap:
            wrapped_lines = []

            if len(self.lines) > 0 and isinstance(self.lines[0], LazilyProcessedTextBuffer):
                if not self.lines[0].processed:
                    self.lines[0].go()
                self.lines = self.lines[0].lines + self.lines[1:]

            for line in self.lines:
                if isinstance(line, Table):
                    wrapped_lines.append(line)
                    self.filled_line_count = -1
                    continue
                if not isinstance(line, AttributedString):
                    line = AttributedString(line)
                max_size = self.width
                indent_size = 10
                indenting = False
                lines = []
                curs = 0

                while True:
                    slice_size = max_size if not indenting else max_size - indent_size
                    slice_size = min(slice_size, len(line.string) - curs)
                    if len(line.string) + 10 - curs > max_size:
                        slice_size = self.find_clean_breakpoint(line.string[curs:curs + slice_size], slice_size)
                    text = line.string[curs:curs + slice_size]
                    if indenting:
                        text = ' ' * 10 + text
                    lines.append(text)
                    curs += slice_size
                    if len(line.string) + 10 - curs <= max_size:
                        text = line.string[curs:]
                        if text.strip() == "":
                            break
                        text = ' ' * 10 + text
                        lines.append(text)
                        break
                    indenting = True

                if lines[-1] == "" and len(lines) > 1:
                    lines.pop()

                if len(line.attrs) > 0:
                    attributes = line.attrs
                    curs = 0
                    alines = []
                    indenting = False
                    for wline in lines:
                        if len(wline) > self.width:
                            global PANIC_STRING
                            PANIC_STRING = "Line wrapping code failed sanity check: String width was larger than window size"
                            raise PanicException
                        attr_str = AttributedString(wline)
                        for attr in attributes:
                            attr_start = max(attr[0][0] - curs, 0)
                            attr_end = attr[0][1] - curs
                            if attr_end > 0:
                                if indenting:
                                    attr_start += 10
                                    attr_end += 10
                                attr_str.set_attr(attr_start, attr_end, attr[1])
                        alines.append(attr_str)
                        curs += len(wline)
                        indenting = True
                    lines = alines

                wrapped_lines += lines

            if not self.filled_line_count == -1:
                self.filled_line_count = len(wrapped_lines)
            self.processed_lines = wrapped_lines

        else:
            trunc_lines = []
            for line in self.lines:
                if not isinstance(line, AttributedString):
                    line = AttributedString(line)
                if len(line.string) > self.width:
                    slice_size = self.width
                    line.string = line.string[0:slice_size - 3] + "..."
                    trunc_lines.append(line)
                else:
                    trunc_lines.append(line)

            self.filled_line_count = len(self.lines)
            self.processed_lines = trunc_lines

    def draw_lines(self):
        """Update the internal representation of lines to be displayed.

        :return:
        """
        x = 0

        display_lines = self.rendered_lines_from(self.processed_lines, self.scrollcursor)
        for y, line in enumerate(display_lines):
            if isinstance(line, AttributedString):
                text = line.string
                self.box.write(x, y, text, self.render_attr)
                for attribute in line.attrs:
                    attr_start = attribute[0][0]
                    if attr_start < 0:
                        continue
                    attr_end = max(attribute[0][1], self.box.width)
                    attr = attribute[1]
                    try:
                        self.box.write(x + attr_start, y, text[attr_start:attr_end - 1], attr)
                    except PanicException:
                        pass
                # DEBUG: self.box.write(x, y, str(line.attrs), self.render_attr)
            else:
                self.box.write(x, y, line.ljust(self.width, ' '), self.render_attr)

    def rendered_lines_from(self, lines, start_line):
        """Return slice of lines based on window height

        :param lines: Full set of lines
        :param start_line: Line start
        :return: Slice of lines
        """
        end_line = start_line + self.height - 1
        pincount = 0
        pins = []

        prop_lines = [*lines]

        for i, line in enumerate(lines):
            if isinstance(line, Table):
                prop_lines = lines[:i]
                table_lines = line.fetch(start_line, int(self.height), self.width).split('\n')
                table_attr_lines = []

                for _line in table_lines:
                    table_attr_lines.append(AttributedString.ansi_to_attrstr(_line))

                prop_lines += table_attr_lines
                start_line = 0
                end_line = self.height - 1

        return pins + prop_lines[start_line + pincount:end_line]


# # # # #
#
# View Level Abstraction:::
# These classes represent (usually) interactive views, all of which sit on/in Display Boxes
#
# # # # #

class View:
    """
    Base View Class - Used for views that don't require too much complex functionality to render.
    """

    def __init__(self):
        self.box = None
        self.children = []
        self.draw = True

    def add_subview(self, view):
        if view not in self.children:
            self.children.append(view)

    def coord_translate(self, x, y):
        return self.box.coord_translate(x, y)

    def redraw(self):
        """
        The .redraw() method is called by the view controller and should be implemented by subclasses to
            re-update the self.box property with its contents. This is how all Views should handle updating/changing
            their content.

        :return:
        """
        pass

    def handle_key_press(self, key):
        """
        This method is called by the View Controller and is passed key-press events.

        key-press events are passed if the View Controller or higher-priority views dont absorb the keypress event.
        If this view decides it should handle it (by returning True), the key-press will be "handled" and not passed
            to any other Views.

        :param key: Key ordinal
        :return: True or False; whether the keypress was handled by this View
        """
        return False

    def handle_mouse(self, x, y):
        """
        This method is called by the View Controller and is passed mouse events.

        Mouse events are passed if the VC itself or higher-priority views dont absorb the mouse event.
        If this view decides it should handle it (by returning True), the mouse event will be "handled" and not passed
            to any other Views.

        :param x: X coordinate of the mouse-press
        :param y: Y coordinate of the mouse-press
        :return: True or False; whether the keypress was handled by this View
        """
        for child in self.children:
            if child.handle_mouse(x, y):
                return True
        return False


class ScrollView(View):
    """
    ScrollView - Used for views which may display text/lists that require scrolling functionality
    """

    def __init__(self):
        super().__init__()
        self.scroll_view = None
        self.scroll_view_text_buffer = None

        # TODO: Implement
        self.scroll_cursor = 0


class Button(View):
    def __init__(self, parent, x, y, height):
        super().__init__()
        self.text = ""
        self.box = Box(parent.box, x, y, 0, height)

    def set_text(self, text):
        self.text = text
        self.box.width = len(text)

    def redraw(self):
        self.box.write(0, 0, self.text, curses.A_NORMAL)

    def handle_mouse(self, x, y):
        tx, ty = self.box.coord_translate(x, y)
        if 0 <= tx <= self.box.width:
            if 0 <= ty <= self.box.height:
                self.action()
                return True
        return False

    def action(self):
        pass


# # # # #
#
# Title Bar:::
# Handles the Title Bar and the Menus it provides
#
# # # # #


class TitleBarMenuItem:
    def __init__(self, text):
        self.text = text
        self.rend_text = f' {text} '
        self.rend_width = len(self.rend_text)
        self.menu_items = []


class HelpMenuItem(TitleBarMenuItem):
    def __init__(self):
        super().__init__("Help")

    def function(self):
        raise HelpMenuException


class FileMenuItem(TitleBarMenuItem):
    def __init__(self):
        super().__init__("File")
        self.menu_items.append(("Open", self.open))
        self.menu_items.append(("Save Edits", self.save))

    def open(self):
        raise FileBrowserOpenNewFileException

    def save(self):
        pass


class EditMenuItem(TitleBarMenuItem):
    def __init__(self):
        super().__init__("Edit")
        self.menu_items.append(("Delete Item", self.delete))

    def delete(self):
        pass


class DumpMenuItem(TitleBarMenuItem):
    def __init__(self):
        super().__init__("Dump")
        self.menu_items.append(("Dump Headers", self.headers))
        self.menu_items.append(("Dump TAPI stub", self.tbd))

    def headers(self):
        pass

    def tbd(self):
        pass


class TitleBar(View):
    """
    The Title Bar represents the top 1st line of the window.
    It's responsible for displaying the window title and menus
    """
    MENUS_START = 10

    def __init__(self):
        super().__init__()

        self.menu_items = []
        self.menu_item_xy_map = {}

        self.pres_menu_item = None
        self.pres_menu_item_index = -1

        self.add_menu_item(HelpMenuItem())

        # self.add_menu_item(FileMenuItem())
        # self.add_menu_item(EditMenuItem())
        # self.add_menu_item(DumpMenuItem())

        self.exit_button = Button(self, 0, 0, 1)
        self.exit_button.set_text(" Exit ")
        self.exit_button.action = self.exit

        self.add_subview(self.exit_button)

    def exit(self):
        raise ExitProgramException

    def add_menu_item(self, item):
        if len(self.menu_items) > 0:
            top_item = self.menu_items[-1]
            start_x = self.menu_item_xy_map[top_item][0] + top_item.rend_width + 2
        else:
            start_x = TitleBar.MENUS_START + 3
        self.menu_items.append(item)
        self.menu_item_xy_map[item] = [start_x, 0]

    def redraw(self):

        if not self.box:
            return

        self.box.write(0, 0, '╒' + '═'.ljust(curses.COLS - 3, '═') + '╕', curses.color_pair(9))
        self.box.write(2, 0, f' {WINDOW_NAME} ', curses.A_NORMAL)
        self.box.write(0, 1, '┟' + ''.ljust(curses.COLS - 3, '━') + '┦', curses.color_pair(9))
        self.box.write(TitleBar.MENUS_START, 0, '╤', curses.color_pair(9))
        self.box.write(TitleBar.MENUS_START, 1, '┸', curses.color_pair(9))

        x = TitleBar.MENUS_START + 3
        for item in self.menu_items:
            self.box.write(x, 0, item.rend_text, curses.A_NORMAL)
            self.box.write(x + 1, 0, item.rend_text[1], curses.A_UNDERLINE)
            x += item.rend_width + 2

        for child in self.children:
            child.redraw()

    def handle_key_press(self, key):
        if self.pres_menu_item_index < 0:
            return False
        if key == curses.KEY_LEFT:
            if self.pres_menu_item_index > 0:
                n_ind = self.pres_menu_item_index - 1
                n_item = self.menu_items[n_ind]
                coords = self.menu_item_xy_map[n_item]
                self.pres_menu_item = (n_item, coords[0])
                self.pres_menu_item_index = n_ind
                raise PresentTitleMenuException

        elif key == curses.KEY_RIGHT:
            if not self.pres_menu_item_index + 1 >= len(self.menu_items):
                n_ind = self.pres_menu_item_index + 1
                n_item = self.menu_items[n_ind]
                coords = self.menu_item_xy_map[n_item]
                self.pres_menu_item = (n_item, coords[0])
                self.pres_menu_item_index = n_ind
                raise PresentTitleMenuException

        return False

    def handle_mouse(self, x, y):
        handle = False

        if y < 1:
            if super().handle_mouse(x, y):
                return True

            handle = True

            if x < 10:
                raise PresentDebugMenuException
            i = 0
            for item, coords in self.menu_item_xy_map.items():
                if coords[0] <= x <= coords[0] + item.rend_width:
                    self.pres_menu_item = (item, coords[0])
                    self.pres_menu_item_index = i
                    raise PresentTitleMenuException
                i += 1

        return handle


# # # # #
#
# Footer Bar:::
# Mainly just holds debug info at the moment. Maybe it will get used later.
#
# # # # #


class FooterBar(View):
    MENUS_START = 40

    def __init__(self):
        super().__init__()
        self.show_debug = False
        self.debug_text = ''

        self.hi_text = ''

        self.now = datetime.now().time()
        graveyard = 'hope your night is going well'
        morning = 'good morning! :)'
        afternoon = 'good afternoon'
        evening = 'good evening ~'
        msgmap = {
            (22, 24): graveyard,
            (0, 4): graveyard,
            (5, 11): morning,
            (12, 16): afternoon,
            (17, 21): evening
        }
        for key, val in msgmap.items():
            if self.now.hour in range(key[0], key[1] + 1):
                self.hi_text = f' {val} '

    def redraw(self):
        self.box.write(0, 0, '╘' + '═'.ljust(curses.COLS - 3, '═') + '╛', curses.color_pair(9))
        self.box.write(FooterBar.MENUS_START, 0, '╧', curses.color_pair(9))

        if self.show_debug:
            self.box.write(self.box.width - len(self.debug_text) - 5, 0, self.debug_text, curses.A_NORMAL)
        else:
            self.box.write(self.box.width - len(self.hi_text) - 5, 0, self.hi_text, curses.color_pair(9))


# # # # #
#
# SideBar:::
# True backbone of the screen; Sidebar is responsible for displaying the various
#   contexts available to be shown on the Main Menu
#
# # # # #


class SidebarMenuItem:
    def __init__(self, name: str, menu_content, parent=None):
        self.name = name
        self.rend_name = name
        self.content = menu_content
        self.parent = parent
        self.children = []
        self.show_children = False
        self.selected = False

    def parse_mmc(self):

        lines = []

        for item in self.content.lines:
            if isinstance(item, str):
                lines += item.split('\n')
            else:
                lines.append(item)

        attrib_content = []
        for item in lines:
            if isinstance(item, str):
                item = item.replace('\t', '    ')  # double tabs break rendering
                attrib_content.append(AttributedString.ansi_to_attrstr(item))
            else:
                attrib_content.append(item)
        self.content = MainMenuContentItem(attrib_content)

    @staticmethod
    def item_list_with_children(menu_item, depth=1):
        """
        Recursive function that returns a single, non-nested, ordered list of items and their children for display.

        :param depth:
        :param menu_item: Root item to recurse through children of.
        :return: List of items generated
        """
        items = [menu_item]
        if menu_item.show_children:
            for child in menu_item.children:
                child.rend_name = '  ' * depth + child.name
                items += SidebarMenuItem.item_list_with_children(child, depth + 1)
        return items


class Sidebar(ScrollView):
    WIDTH = SIDEBAR_WIDTH

    def __init__(self):
        super().__init__()

        self.selected_index = 0

        self.current_sidebar_item_count = 0
        self.processed_items = []

        self.items = []

    def redraw(self):
        # Redraw right-side divider
        self.update_item_listing()
        for index in range(0, curses.LINES - 2):
            self.box.write(0, index, VERT_LINE, curses.color_pair(9))
            self.box.write(Sidebar.WIDTH, index, VERT_LINE, curses.color_pair(9))
        self.box.write(SIDEBAR_WIDTH, 0, '┰', curses.color_pair(9))

    def select_item(self, index):
        """
        Update selected item index and redraw the item listing.

        :param index:
        :return:
        """
        self.selected_index = index
        self.update_item_listing()

    def update_item_listing(self):
        """
        Re-process the entire item listing to generate the content to be displayed.

        :return:
        """

        self.processed_items = []
        for item in self.items:
            self.processed_items += SidebarMenuItem.item_list_with_children(item)

        self.current_sidebar_item_count = len(self.processed_items)

        self.scroll_view_text_buffer.lines = []

        for index, item in enumerate(self.processed_items):
            name = item.rend_name.ljust(SIDEBAR_WIDTH - 6, ' ')

            if len(item.children) > 0:
                if item.show_children:
                    name = name + '-'
                else:
                    name = name + '+'
            else:
                name = name + ' '
            if self.selected_index == index:
                name = AttributedString(name)
                name.set_attr(0, len(name.string) - 1, Attribute.HIGHLIGHTED)
                self.scroll_view_text_buffer.lines.append(name)
            else:
                self.scroll_view_text_buffer.lines.append(name)

        # This chunk of code makes sure the currently selected sidebar position is always visible.
        # It scrolls as the user selection moves up/down
        # The "while" loop makes sure that, if a submenu multiple times the size of the screen height was collapsed,
        #       the scroll cursor will jump to a position where things are visible again.
        scroll_view_height = self.scroll_view_text_buffer.height
        if self.selected_index - scroll_view_height + 4 > self.scroll_view_text_buffer.scrollcursor:
            while self.selected_index - scroll_view_height + 4 > self.scroll_view_text_buffer.scrollcursor:
                self.scroll_view_text_buffer.scrollcursor += 1
        elif self.selected_index < self.scroll_view_text_buffer.scrollcursor + 2 and self.scroll_view_text_buffer.scrollcursor > 0:
            while self.selected_index < self.scroll_view_text_buffer.scrollcursor + 2 and self.scroll_view_text_buffer.scrollcursor > 0:
                self.scroll_view_text_buffer.scrollcursor -= 1

        self.scroll_view_text_buffer.process_lines()
        self.scroll_view_text_buffer.draw_lines()

    def add_menu_item(self, item):
        """
        Add a root menu item to the Sidebar and update the list.

        :param item:
        :return:
        """
        self.items.append(item)
        self.update_item_listing()

    def collapse_index(self, index):
        """
        Collapse item at index if it is open.

        :param index:
        :return:
        """
        item = self.processed_items[index]
        if len(item.children) > 0 and item.show_children:
            item.show_children = False

            self.update_item_listing()
        else:
            swap_to_index = 0
            for index, pitem in enumerate(self.processed_items):
                if pitem == item.parent:
                    swap_to_index = index
                    break
            pitem = self.processed_items[swap_to_index]
            pitem.show_children = False

            self.update_item_listing()
            self.select_item(swap_to_index)
            self.update_item_listing()

    def handle_key_press(self, key):
        if key == curses.KEY_UP:
            index = self.selected_index
            if index - 1 in range(0, self.current_sidebar_item_count):
                self.select_item(index - 1)

        elif key == curses.KEY_DOWN:
            index = self.selected_index
            if index + 1 in range(0, self.current_sidebar_item_count):
                self.select_item(index + 1)

        elif key == curses.KEY_LEFT:
            index = self.selected_index
            self.collapse_index(index)

        elif key == curses.KEY_RIGHT:
            index = self.selected_index
            item = self.processed_items[index]
            if len(item.children) > 0:
                item.show_children = True
                self.update_item_listing()

        elif key == ord(" "):
            index = self.selected_index
            item = self.processed_items[index]
            if item.show_children:
                self.collapse_index(index)
            else:
                if len(item.children) > 0:
                    item.show_children = True
                    self.update_item_listing()
        else:
            return False

        return True

    def handle_mouse(self, x, y):
        absorb = False
        if x < SIDEBAR_WIDTH and y > self.box.y:
            absorb = True
            y = y - 3
            x = x - 2

            if -1 <= x - SIDEBAR_WIDTH + 6 <= 1 and y < len(self.processed_items):
                index = y
                item = self.processed_items[index]
                if len(item.children) > 0:
                    if item.show_children:
                        self.selected_index = y
                        self.collapse_index(y)
                    else:
                        item.show_children = True
                        self.update_item_listing()
            elif y < len(self.processed_items):
                self.select_item(y)

        return absorb


# # # # #
#
# Main Screen:::
# Displays text from the current context.
#
# # # # #


class MainMenuContentItem:
    """
    Just holds the unprocessed lines to be displayed in the Main Menu.

    These are generated at the start and stored in Sidebar Menu Items, from which the View Controller pulls them out and
        displays the in the Main Screen.
    """

    def __init__(self, lines=None):
        if lines is None:
            lines = []
        self.lines = lines


class MainScreen(ScrollView):
    """
    Displays the currently relevant text from the context.
    """

    def __init__(self):
        super().__init__()
        self.info_box = None

        self.tabname = ""
        self.highlighted = False

        self.currently_displayed_index = 0

    def set_tab_name(self, name):
        """
        Update the tab name

        :param name:
        :return:
        """
        self.tabname = name

    def redraw(self):
        width = curses.COLS - Sidebar.WIDTH - 2

        for index in range(0, curses.LINES - 2):
            self.box.write(width, index, VERT_LINE, curses.color_pair(9))

        # Redraw the text
        self.scroll_view_text_buffer.draw_lines()

        width = curses.COLS - Sidebar.WIDTH - 3
        # Clear and Redraw the Info Box bar
        self.info_box.write(-1, 1, '╞' + ''.ljust(width, '═') + '╡', curses.color_pair(9))
        self.info_box.write(2, 1, f'╡ {self.tabname.strip()} ╞', curses.color_pair(9))
        self.info_box.write(3, 1, f' {self.tabname.strip()} ',
                            curses.A_NORMAL if not self.highlighted else curses.A_STANDOUT)

    def handle_key_press(self, key):
        if key == curses.KEY_UP:
            self.scroll_view_text_buffer.scrollcursor = max(0, self.scroll_view_text_buffer.scrollcursor - 1)
            self.scroll_view_text_buffer.draw_lines()
            return True
        elif key == curses.KEY_PPAGE:
            j_height = self.scroll_view_text_buffer.box.height - 3
            for i in range(j_height):
                self.scroll_view_text_buffer.scrollcursor = max(0, self.scroll_view_text_buffer.scrollcursor - 1)

            self.scroll_view_text_buffer.draw_lines()
            return True

        elif key == curses.KEY_DOWN:
            if self.scroll_view_text_buffer.filled_line_count == -1:
                self.scroll_view_text_buffer.scrollcursor += 1
            else:
                self.scroll_view_text_buffer.scrollcursor = min(
                    self.scroll_view_text_buffer.filled_line_count - self.scroll_view_text_buffer.height + 1,
                    self.scroll_view_text_buffer.scrollcursor + 1)
            self.scroll_view_text_buffer.draw_lines()
            return True

        elif key == curses.KEY_NPAGE:
            j_height = self.scroll_view_text_buffer.box.height - 3
            for i in range(j_height):
                if self.scroll_view_text_buffer.filled_line_count == -1:
                    self.scroll_view_text_buffer.scrollcursor += 1
                else:
                    self.scroll_view_text_buffer.scrollcursor = min(
                        self.scroll_view_text_buffer.filled_line_count - self.scroll_view_text_buffer.height + 1,
                        self.scroll_view_text_buffer.scrollcursor + 1)
            self.scroll_view_text_buffer.draw_lines()
            return True
        elif key == ord("d"):
            global ATTR_STRING_DEBUG
            ATTR_STRING_DEBUG = True
            raise RebuildAllException

        return False


class DebugMenu(ScrollView):
    def __init__(self):
        super().__init__()
        self.draw = False

    def parse_lines(self):
        attrib_content = []
        for item in self.scroll_view_text_buffer.lines:
            if isinstance(item, str):
                attrib_content.append(AttributedString.ansi_to_attrstr(item))
            else:
                attrib_content.append(item)

        self.scroll_view_text_buffer.lines = attrib_content

    def redraw(self):
        width = self.box.width - 10
        height = self.box.height
        lc = '╒'
        rc = '╕'
        div = '═'
        bl = '╘'
        br = '╛'

        bgcolor = curses.color_pair(1)

        self.box.write(0, 0, lc + ''.ljust(width - 2, div) + rc, bgcolor)

        for line in range(1, height):
            self.box.write(0, line, VERT_LINE + ''.ljust(width - 2, ' ') + VERT_LINE, bgcolor)

        self.box.write(0, height, bl + ''.ljust(width - 2, div) + br, bgcolor)

        self.scroll_view_text_buffer.process_lines()
        self.scroll_view_text_buffer.draw_lines()

    def handle_key_press(self, key):
        if key == curses.KEY_UP:
            self.scroll_view_text_buffer.scrollcursor = max(0, self.scroll_view_text_buffer.scrollcursor - 1)
            self.scroll_view_text_buffer.draw_lines()
            return True
        elif key == curses.KEY_DOWN:
            self.scroll_view_text_buffer.scrollcursor = min(
                self.scroll_view_text_buffer.filled_line_count - self.scroll_view_text_buffer.height + 1,
                self.scroll_view_text_buffer.scrollcursor + 1)
            self.scroll_view_text_buffer.draw_lines()
            return True
        return False

    def handle_mouse(self, x, y):

        x = x - self.box.x
        y = y - self.box.y

        handle = False
        if not self.draw:
            return handle

        if y == 0:
            handle = True
            if -1 <= x - self.box.width + 15 <= 1:
                self.draw = False

        return handle


class HelpMenu(ScrollView):
    def __init__(self):
        super().__init__()
        self.draw = False

    def parse_lines(self):
        attrib_content = []
        for item in self.scroll_view_text_buffer.lines:
            if isinstance(item, str):
                attrib_content.append(AttributedString.ansi_to_attrstr(item))
            else:
                attrib_content.append(item)

        self.scroll_view_text_buffer.lines = attrib_content

    def redraw(self):
        width = self.box.width - 10
        height = self.box.height
        lc = '╒'
        rc = '╕'
        div = '═'
        bl = '╘'
        br = '╛'

        bgcolor = curses.color_pair(1)

        self.box.write(0, 0, lc + ''.ljust(width - 2, div) + rc, bgcolor)

        for line in range(1, height):
            self.box.write(0, line, VERT_LINE + ''.ljust(width - 2, ' ') + VERT_LINE, bgcolor)

        self.box.write(0, height, bl + ''.ljust(width - 2, div) + br, bgcolor)

        self.scroll_view_text_buffer.process_lines()
        self.scroll_view_text_buffer.draw_lines()

    def handle_key_press(self, key):
        if key == curses.KEY_UP:
            self.scroll_view_text_buffer.scrollcursor = max(0, self.scroll_view_text_buffer.scrollcursor - 1)
            self.scroll_view_text_buffer.draw_lines()
            return True
        elif key == curses.KEY_DOWN:
            self.scroll_view_text_buffer.scrollcursor = min(
                self.scroll_view_text_buffer.filled_line_count - self.scroll_view_text_buffer.height + 1,
                self.scroll_view_text_buffer.scrollcursor + 1)
            self.scroll_view_text_buffer.draw_lines()
            return True
        return False

    def handle_mouse(self, x, y):

        x = x - self.box.x
        y = y - self.box.y

        handle = False
        if not self.draw:
            return handle

        if x < 0 or x > self.box.width or y < 0 or y > self.box.height:
            self.draw = False
            handle=True

        if y == 0:
            handle = True
            if -1 <= x - self.box.width + 15 <= 1:
                self.draw = False

        return handle


class LoaderStatusView(View):
    def __init__(self):
        super().__init__()
        self.draw = False
        self.status_string = "Loading..."

    def redraw(self):
        if not self.draw:
            return
        box_width = max(40, len(self.status_string) + 10)
        start_x = ceil(self.box.width / 2 - box_width / 2)
        height = 4
        start_y = ceil(self.box.height / 2 - height / 2)

        for i in range(start_y, start_y + height + 1):
            self.box.write(start_x, i, ' ' * box_width, curses.A_STANDOUT)

        i = 0
        for line in self.status_string.split('\n'):
            self.box.write(start_x + 1, start_y + 1 + i, line, curses.A_STANDOUT)
            i += 1


class UserInputPrompt(View):
    def __init__(self):
        super().__init__()
        self.draw = False
        self.prompt_string = ""
        self.user_input_is_string = False

        self.active_render_subbox = None
        self.response = None

    def redraw(self):
        if not self.draw:
            return
        box_width = max(40, len(self.prompt_string) + 10)
        start_x = ceil(self.box.width / 2 - box_width / 2)
        height = 4
        start_y = ceil(self.box.height / 2 - height / 2)

        self.active_render_subbox = Box(None, start_x, start_y, box_width, height)

        for i in range(start_y, start_y + height + 1):
            self.box.write(start_x, i, ' ' * box_width, curses.A_STANDOUT)

        xof = 2
        for msg in ['YES', 'NO', 'CANCEL']:
            self.box.write(xof, height, msg, curses.A_STANDOUT)
            xof += len(msg) + 2

    def handle_mouse(self, x, y):
        if not self.draw:
            return False

        if self.active_render_subbox.is_click_inbounds(x, y) and y - self.active_render_subbox.y == 4:
            xof = 2
            xp = x - self.active_render_subbox.x
            height = 4
            for msg in ['YES', 'NO', 'CANCEL']:
                self.box.write(xof, height, msg, curses.A_STANDOUT)
                wid = len(msg)
                if xp in range(xof, xof + wid + 1):
                    if msg == 'YES':
                        self.response = 'y'
                    elif msg == 'NO':
                        self.response = 'n'
                    else:
                        self.response = 'c'
                    return True
                xof += len(msg) + 2

        return False


class MenuOverlayRenderingView(View):
    def __init__(self):
        super().__init__()
        self.draw = False
        self.active_render_menu = None
        self.active_menu_start_x = 0

        self.active_render_subbox = None

    def redraw(self):
        if not self.draw or not self.active_render_menu:
            return

        if len(self.active_render_menu.menu_items) == 0:
            self.active_render_menu.function()
            self.draw = False

        start = (self.active_menu_start_x + 1, 1)
        width = 30
        height = len(self.active_render_menu.menu_items) + 2

        self.active_render_subbox = Box(None, start[0], start[1], width, height)

        for line in range(start[1], start[1] + height):
            self.box.write(start[0], line, ' ' * width, curses.A_STANDOUT)

        for linen, item in enumerate([i[0] for i in self.active_render_menu.menu_items]):
            self.box.write(start[0] + 1, start[1] + 1 + linen, f'{item}', curses.A_STANDOUT)

    def handle_mouse(self, x, y):
        if not self.draw or not self.active_render_menu:
            return False

        if self.active_render_subbox.is_click_inbounds(x, y):
            # x = x - self.active_render_subbox.x
            y = y - self.active_render_subbox.y - 1
            if y < 0:
                return False
            if y > len(self.active_render_menu.menu_items):
                raise DestroyTitleMenuException
            if y == len(self.active_render_menu.menu_items):
                return False
            item_tup = self.active_render_menu.menu_items[y]
            item_tup[1]()
            return True

        raise DestroyTitleMenuException


class FileSystemBrowserOverlayView(ScrollView):
    def __init__(self):
        super().__init__()
        self.draw = False
        self.current_dir_path = os.getcwd()
        self.select_file = True
        self.callback = None

        self.selected_index = 0

    def redraw(self):
        start_x = 3
        width = self.box.width - (2 * start_x)
        start_y = start_x
        height = self.box.height - (2 * start_y)

        lc = '╒'
        rc = '╕'
        div = '═'
        bl = '╘'
        br = '╛'
        lj = '╞'
        rj = '╡'

        bgcolor = curses.A_NORMAL

        self.box.write(start_x, start_y, lc + ''.ljust(width - 2, div) + rc, bgcolor)

        for line in range(start_y + 1, start_y + height + 1):
            self.box.write(start_x, line, VERT_LINE + ''.ljust(width - 2, ' ') + VERT_LINE, bgcolor)

        self.box.write(start_x, start_y + height, bl + ''.ljust(width - 2, div) + br, bgcolor)
        self.box.write(start_x, start_y + height - 2, lj + ''.ljust(width - 2, div) + rj, bgcolor)
        self.box.write(start_x + width - 10, start_y + height - 1, ' SELECT ', curses.A_STANDOUT)

        lines = ['..'] + [i for i in os.listdir(self.current_dir_path)]
        nlines = []
        for item in lines:
            if os.path.isdir(item):
                nlines.append(f'{item}/')
            else:
                nlines.append(item)

        for index, name in enumerate(nlines):
            if self.selected_index == index:
                name = AttributedString(name)
                name.set_attr(0, len(name.string) - 1, Attribute.HIGHLIGHTED)
                self.scroll_view_text_buffer.lines.append(name)
            else:
                self.scroll_view_text_buffer.lines.append(name)

        # This chunk of code makes sure the currently selected sidebar position is always visible.
        # It scrolls as the user selection moves up/down
        # The "while" loop makes sure that, if a submenu multiple times the size of the screen height was collapsed,
        #       the scroll cursor will jump to a position where things are visible again.
        scroll_view_height = self.scroll_view_text_buffer.height
        if self.selected_index - scroll_view_height + 4 > self.scroll_view_text_buffer.scrollcursor:
            while self.selected_index - scroll_view_height + 4 > self.scroll_view_text_buffer.scrollcursor:
                self.scroll_view_text_buffer.scrollcursor += 1
        elif self.selected_index < self.scroll_view_text_buffer.scrollcursor + 2 and self.scroll_view_text_buffer.scrollcursor > 0:
            while self.selected_index < self.scroll_view_text_buffer.scrollcursor + 2 and self.scroll_view_text_buffer.scrollcursor > 0:
                self.scroll_view_text_buffer.scrollcursor -= 1

        self.scroll_view_text_buffer.process_lines()
        self.scroll_view_text_buffer.draw_lines()

    def handle_key_press(self, key):
        if not self.draw:
            return False

        if key == curses.KEY_UP:
            index = self.selected_index
            if index - 1 in range(0, len(self.scroll_view_text_buffer.lines)):
                self.selected_index -= 1
            return True

        elif key == curses.KEY_DOWN:
            index = self.selected_index
            if index + 1 in range(0, len(self.scroll_view_text_buffer.lines)):
                self.selected_index += 1
            return True

        return False


# # # # #
#
# File Loaders:::
# Handles creating Sidebar and MainMenuContent items for a given file
#
# # # # #


class KToolMachOLoader:
    SUPPORTS_256 = False
    SUPPORTS_COLOR = True
    HARD_FAIL = False
    CUR_SL = 0
    SL_CNT = 0

    @staticmethod
    def parent_count(item):
        count = 0
        item = item
        while item.parent is not None:
            count += 1
            item = item.parent
        return count

    @classmethod
    def contents_for_file(cls, fd, callback, mmap=True):
        machofile = MachOFile(fd, use_mmaped_io=mmap)
        items = []
        KToolMachOLoader.SL_CNT = len(machofile.slices)
        for macho_slice in machofile.slices:
            KToolMachOLoader.CUR_SL += 1
            try:
                items.append(cls.slice_item(macho_slice, callback))
            except Exception as ex:
                raise ex
        return items

    @staticmethod
    def slice_item(macho_slice, callback):
        loaded_image = Dyld.load(macho_slice)
        if hasattr(macho_slice, 'type'):
            slice_nick = f'{macho_slice.type.name}:{macho_slice.subtype.name}' + " Slice"
        else:
            slice_nick = "Thin MachO"
        callback(f'Slice {KToolMachOLoader.CUR_SL}/{KToolMachOLoader.SL_CNT}\nLoading MachO Image')
        slice_item = SidebarMenuItem(f'{slice_nick}', None, None)
        slice_item.content = KToolMachOLoader._file(loaded_image, slice_item, callback).content
        items = [KToolMachOLoader.load_cmds,
                 KToolMachOLoader.segments,
                 KToolMachOLoader.codesign,
                 KToolMachOLoader.linked,
                 KToolMachOLoader.imports,
                 KToolMachOLoader.exports,
                 KToolMachOLoader.symtab]
        for item in items:
            try:
                slice_item.children.append(item(loaded_image, slice_item, callback))
            except Exception as ex:
                if KToolMachOLoader.HARD_FAIL:
                    raise ex
                else:
                    pass
        try:
            slice_item.children += KToolMachOLoader.objc_items(loaded_image, slice_item, callback)
        except Exception as ex:
            if KToolMachOLoader.HARD_FAIL:
                raise ex
            else:
                pass

        try:
            slice_item.children += KToolMachOLoader.swift_items(loaded_image, slice_item, callback)
        except Exception as ex:
            if KToolMachOLoader.HARD_FAIL:
                pass
            else:
                pass
        slice_item.show_children = True
        return slice_item

    @staticmethod
    def segments(lib, parent=None, callback=None):
        callback(f'Slice {KToolMachOLoader.CUR_SL}/{KToolMachOLoader.SL_CNT}\nLoading Segments & Generating Hexdumps')
        smmci = MainMenuContentItem()
        ssmi = SidebarMenuItem("Segments", smmci, parent)
        table = Table()
        table.titles = ['Segment Name', 'VM Address', 'Size', 'File Address']

        for segname, segm in lib.segments.items():

            table.rows.append([segname, hex(segm.vm_address), hex(segm.size), hex(segm.file_address)])

            segtable = Table()
            segtable.titles = ['Section Name', 'VM Address', 'Size', 'File Address']
            mmci = MainMenuContentItem()
            item = SidebarMenuItem(f'{segname} [{len(segm.sections.items())} Sections]', mmci, ssmi)
            for secname, sect in segm.sections.items():
                segtable.rows.append([secname, hex(sect.vm_address), hex(sect.size), hex(sect.file_address)])
                hextab = HexDumpTable()
                hextab.hex = lib.slice.get_bytes_at(sect.file_address, sect.size)
                itm = MainMenuContentItem()
                itm.lines.append(hextab)
                item.children.append(SidebarMenuItem(secname, itm, item))
            ssmi.children.append(item)
            if len(segm.sections.items()) == 0:
                if 'PAGEZERO' not in segname:
                    hextab = HexDumpTable()
                    hextab.hex = lib.slice.get_bytes_at(segm.file_address, segm.size)
                    mmci.lines.append(hextab)
            else:
                mmci.lines.append(segtable)

        smmci.lines.append(table)

        ssmi.parse_mmc()
        return ssmi

    # noinspection PyUnusedLocal
    @staticmethod
    def _file(lib, parent=None, callback=None):
        file_content_item = MainMenuContentItem()

        file_content_item.lines.append(f'Install Name: §35m{lib.install_name}§39m')
        file_content_item.lines.append(f'Filetype: §35m{lib.macho_header.filetype.name}§39m')
        file_content_item.lines.append(f'Flags: §35m{"§39m, §35m".join([i.name for i in lib.macho_header.flags])}§39m')
        if lib.uuid:
            file_content_item.lines.append(f'UUID: §35m{lib.uuid.hex().upper()}§39m')
        file_content_item.lines.append(f'Platform: §35m{lib.platform.name}§39m')
        file_content_item.lines.append(f'Minimum OS: §35m{lib.minos.x}.{lib.minos.y}.{lib.minos.z}§39m')
        file_content_item.lines.append(
            f'SDK Version: §35m{lib.sdk_version.x}.{lib.sdk_version.y}.{lib.sdk_version.z}§39m')

        file_content_item.lines.append('')
        file_content_item.lines.append(f'macho_header: {str(lib.macho_header.dyld_header)}')

        menuitem = SidebarMenuItem("File Info", file_content_item, parent)
        menuitem.parse_mmc()

        return menuitem

    @staticmethod
    def linked(lib, parent=None, callback=None):

        callback(f'Slice {KToolMachOLoader.CUR_SL}/{KToolMachOLoader.SL_CNT}\nLoading Linked Libs')

        linked_libs_item = MainMenuContentItem()
        for exlib in lib.linked_images:
            linked_libs_item.lines.append(
                '§31m(Weak)§39m ' + exlib.install_name if exlib.weak else '' + exlib.install_name)

        menuitem = SidebarMenuItem("Linked Libraries", linked_libs_item, parent)
        menuitem.parse_mmc()
        return menuitem

    @staticmethod
    def codesign(lib, parent=None, callback=None):
        callback(f'Codesign Info')

        cs = MainMenuContentItem()

        image = lib

        lines = []
        lines += image.codesign_info.superblob.render_indented().split('\n')
        lines.append('')
        for slot in image.codesign_info.slots:
            lines += slot.render_indented().split('\n')
            lines.append('')

        cs.lines = lines

        smi = SidebarMenuItem("Codesign Info", cs, parent)

        # ent item
        ent_mmci = MainMenuContentItem()
        ent_mmci.lines.append(image.codesign_info.entitlements)
        ent_smi = SidebarMenuItem("Entitlements", ent_mmci, smi)
        ent_smi.parse_mmc()
        smi.children.append(ent_smi)

        # req item
        req_mmci = MainMenuContentItem()

        hexdump = HexDumpTable()
        hexdump.hex = bytearray(image.codesign_info.req_dat)

        req_mmci.lines.append(hexdump)

        req_smi = SidebarMenuItem("Requirements", req_mmci, smi)
        smi.children.append(req_smi)

        smi.parse_mmc()
        return smi

    @staticmethod
    def load_cmds(lib, parent=None, callback=None):

        callback(f'Slice {KToolMachOLoader.CUR_SL}/{KToolMachOLoader.SL_CNT}\nProcessing Load Commands')

        load_cmds = MainMenuContentItem()

        lines = [f'Load Command Count: {len(lib.macho_header.load_commands)}']

        load_cmds.lines = lines

        menuitem = SidebarMenuItem("Load Commands", load_cmds, parent)

        for cmd in lib.macho_header.load_commands:
            mmci = MainMenuContentItem()
            mmci.lines = cmd.render_indented().split('\n')
            mmci.lines.append('')
            raw: bytes = lib.slice.get_bytes_at(cmd.off, cmd.cmdsize)
            hexdump = HexDumpTable()
            hexdump.hex = bytearray(raw)

            # noinspection PyTypeChecker
            mmci.lines.append(hexdump)

            try:
                item_name = LOAD_COMMAND(cmd.cmd).name
            except ValueError:
                item_name = str(cmd.cmd)

            lc_menu_item = SidebarMenuItem(item_name, mmci, menuitem)
            menuitem.children.append(lc_menu_item)

        menuitem.parse_mmc()
        return menuitem

    @staticmethod
    def symtab(lib, parent=None, callback=None):
        callback(f'Slice {KToolMachOLoader.CUR_SL}/{KToolMachOLoader.SL_CNT}\nProcessing Symtab')
        mmci = MainMenuContentItem()
        tab = Table()
        tab.titles = ['Address', 'Name']
        for sym in lib.symbol_table.table:
            tab.rows.append([hex(sym.address), sym.fullname])
        mmci.lines.append(tab)

        menuitem = SidebarMenuItem("Symbol Table", mmci, parent)

        menuitem.parse_mmc()
        return menuitem

    # noinspection PyUnusedLocal
    @staticmethod
    def vm_map(lib, parent=None, callback=None):
        mmci = MainMenuContentItem()

        mmci.lines = str(lib.vm).split('\n')

        menuitem = SidebarMenuItem("VM Memory Map", mmci, parent)

        menuitem.parse_mmc()
        return menuitem

    @staticmethod
    def objc_items(lib, parent=None, callback=None):
        objc_lib = ObjCImage.from_image(lib)

        return [KToolMachOLoader.objc_headers(objc_lib, parent, callback)]

    @staticmethod
    def swift_items(lib, parent=None, callback=None):
        return [KToolMachOLoader.swift_types(lib, parent, callback)]

    @staticmethod
    def swift_types(lib, parent=None, callback=None):
        callback(f'Loading Swift Types')
        root_mmci = MainMenuContentItem()

        types = load_swift_types(lib)

        root_mmci.lines = [f'{i.name}::{i.__class__.__name__}' for i in types]

        smi = SidebarMenuItem("Swift Types", root_mmci, parent)

        for _type in types:
            mmci = MainMenuContentItem()
            mmci.lines.append(f'{_type.__class__.__name__}')
            mmci.lines.append(f'')
            if isinstance(_type, SwiftClass):
                mmci.lines.append(str(_type.class_desc))
                mmci.lines.append(f'')
                mmci.lines.append(str([str(i) for i in _type.field_desc.fields]))
                mmci.lines.append(f'')
                mmci.lines.append(str([str(i) for i in _type.ivars]))
                mmci.lines.append(f'')
            else:
                mmci.lines.append(str(_type.typedesc))
                mmci.lines.append('')
                mmci.lines.append(str(_type.field_desc.desc))
                mmci.lines.append(f'')
                mmci.lines.append(str([str(i) for i in _type.field_desc.fields]))
                mmci.lines.append(f'')

            ssmi = SidebarMenuItem(_type.name, mmci, smi)

            smi.children.append(ssmi)

        smi.parse_mmc()
        return smi

    @staticmethod
    def imports(lib, parent=None, callback=None):
        callback(f'Slice {KToolMachOLoader.CUR_SL}/{KToolMachOLoader.SL_CNT}\nProcessing Imports')
        mmci = MainMenuContentItem()

        table = Table()
        table.titles = ['Address', 'Symbol', 'Library']

        for symbol in lib.imports:
            table.rows.append(
                [hex(symbol.address), symbol.fullname, lib.linked_images[int(symbol.ordinal) - 1].install_name])
        mmci.lines.append(table)

        menuitem = SidebarMenuItem("Imports", mmci, parent)

        menuitem.parse_mmc()
        return menuitem

    @staticmethod
    def exports(lib, parent=None, callback=None):
        callback(f'Slice {KToolMachOLoader.CUR_SL}/{KToolMachOLoader.SL_CNT}\nProcessing Exports')
        mmci = MainMenuContentItem()

        table = Table()
        table.titles = ['Address', 'Symbol']

        for symbol in lib.exports:
            table.rows.append([hex(symbol.address), symbol.fullname])

        mmci.lines.append(table)

        menuitem = SidebarMenuItem("Exports", mmci, parent)

        menuitem.parse_mmc()
        return menuitem

    @staticmethod
    def get_header_item(text, name):
        mmci = MainMenuContentItem()
        formatter = Terminal256Formatter() if KToolMachOLoader.SUPPORTS_256 else TerminalFormatter()
        text = highlight(text, ObjectiveCLexer(), formatter)
        lines = text.split('\n')
        mmci.lines = lines
        h_menu_item = SidebarMenuItem(name, mmci, None)
        return h_menu_item

    @staticmethod
    def get_header_text(text):
        formatter = Terminal256Formatter() if KToolMachOLoader.SUPPORTS_256 else TerminalFormatter()
        text = highlight(text, ObjectiveCLexer(), formatter)

        lines = []

        for item in text.split('\n'):
            lines.append(AttributedString.ansi_to_attrstr(item))

        return lines

    @staticmethod
    def objc_headers(objc_lib, parent=None, callback=None):
        generator = HeaderGenerator(objc_lib)
        hnci = MainMenuContentItem()
        hnci.lines = generator.headers.keys()
        menuitem = SidebarMenuItem("ObjC Headers", hnci, parent)
        count = len(generator.headers.keys())
        callback(
            f'Slice {KToolMachOLoader.CUR_SL}/{KToolMachOLoader.SL_CNT}\nProcessing {count} ObjC Headers')

        items = []
        i = 1
        for header_name, header in generator.headers.items():
            callback(
                f'Slice {KToolMachOLoader.CUR_SL}/{KToolMachOLoader.SL_CNT}\nProcessing {count} ObjC Headers\n{i}/{count}')
            mmci = MainMenuContentItem()
            buffer = LazilyProcessedTextBuffer()
            buffer.target = KToolMachOLoader.get_header_text
            buffer.target_args = [str(header)]
            mmci.lines = [buffer]
            sbmi = SidebarMenuItem(header_name, mmci, menuitem)
            items.append(sbmi)
            i += 1

        menuitem.children = items

        menuitem.parse_mmc()
        return menuitem


class KToolKernelCacheLoader(KToolMachOLoader):

    @staticmethod
    def slice_item(macho_slice, callback):
        loaded_image = Dyld.load(macho_slice)
        slice_nick = f'Kernel Cache'
        callback(f'Kernel Cache\nLoading MachO Image')
        slice_item = SidebarMenuItem(f'Kernel Cache', None, None)
        kcache = KernelCache(macho_slice.macho_file)
        slice_item.content = KToolKernelCacheLoader._file(kcache, slice_item, callback).content
        items = [KToolMachOLoader.load_cmds,
                 KToolMachOLoader.segments,
                 KToolMachOLoader.symtab]

        for item in items:
            try:
                slice_item.children.append(item(loaded_image, slice_item, callback))
            except Exception as ex:
                if KToolMachOLoader.HARD_FAIL:
                    raise ex
                else:
                    pass

        kernel_items = [KToolKernelCacheLoader.get_kexts]

        for item in kernel_items:
            try:
                slice_item.children.append(item(kcache, slice_item, callback))
            except Exception as ex:
                if KToolMachOLoader.HARD_FAIL:
                    raise ex
                else:
                    pass

        slice_item.show_children = True
        return slice_item

    @staticmethod
    def _file(kcache, parent=None, callback=None):
        file_content_item = MainMenuContentItem()
        lib = kcache.mach_kernel

        file_content_item.lines.append(f'Release Type: §35m{kcache.release_type}§39m')
        file_content_item.lines.append(f'SOC: §35m{kcache.soc}§39m')
        file_content_item.lines.append(f'Arch: §35m{kcache.arch}§39m')

        file_content_item.lines.append('')

        file_content_item.lines.append(f'Kernel Version: §35m{kcache.version}§39m')
        file_content_item.lines.append(f'{kcache.version_str}')
        if lib.uuid:
            file_content_item.lines.append(f'UUID: §35m{lib.uuid.hex().upper()}§39m')
        file_content_item.lines.append(f'Platform: §35m{lib.platform.name}§39m')

        file_content_item.lines.append(f'Minimum OS: §35m{lib.minos.x}.{lib.minos.y}.{lib.minos.z}§39m')
        file_content_item.lines.append(
            f'SDK Version: §35m{lib.sdk_version.x}.{lib.sdk_version.y}.{lib.sdk_version.z}§39m')

        menuitem = SidebarMenuItem("File Info", file_content_item, parent)
        menuitem.parse_mmc()

        return menuitem

    @staticmethod
    def get_kexts(kcache: KernelCache, parent=None, callback=None):

        hnci = MainMenuContentItem()
        hnci.lines = [kext.name for kext in kcache.kexts]
        menuitem = SidebarMenuItem("KEXTs", hnci, parent)
        count = len(kcache.kexts)
        callback(
            f'Kernel Cache\nProcessing {count} KEXTs')

        items = []
        i = 1
        for kext in kcache.kexts:
            callback(
                f'Kernel Cache\nProcessing {count} KEXT \n{i}/{count}')
            mmci = MainMenuContentItem()
            mmci.lines = []
            mmci.lines.append(f'ID: {kext.name}')
            mmci.lines.append(f'Embedded Version: {kext.version}')

            if kext.prelink_info:
                bundle_text = f"Executable Name: {kext.executable_name}\n{kext.info_string}\nVersion: {kext.version_str}\nStart Address: {hex(kext.start_addr | (0xffff000000000000 if kcache.mach_kernel.vm.detag_kern_64 else 0))}".split('\n')
                bundle_text += ['', '']

                bundle_text += pprint.pformat(kext.prelink_info).split('\n')

                mmci.lines += bundle_text
            sbmi = SidebarMenuItem(kext.name, mmci, menuitem)
            items.append(sbmi)
            i += 1

        menuitem.children = items

        menuitem.parse_mmc()
        return menuitem


# # # # #
#
# Main Screen:::
# Sets up, tears down, rebuilds, redraws, and controls the various views.
# Houses the Main run-loop.
# Performs root level exception handling
# Arbitrates Mouse/Keyboard events throughout the program.
#
# # # # #


class KToolScreen:
    def __init__(self, hard_fail=False):

        self.supports_color = False
        self.supported_colors = 0
        self.hard_fail = hard_fail
        KToolMachOLoader.HARD_FAIL = hard_fail

        self.stdscr = self.setup()

        self.root = RootBox(self.stdscr)

        self.filename = ""

        self.titlebar = TitleBar()
        self.sidebar = Sidebar()
        self.mainscreen = MainScreen()
        self.footerbar = FooterBar()
        self.debug_menu = DebugMenu()
        self.help_menu = HelpMenu()

        self.title_menu_overlay = MenuOverlayRenderingView()
        self.loader_status = LoaderStatusView()
        self.file_browser = FileSystemBrowserOverlayView()
        self.input_overlay = UserInputPrompt()

        self.is_showing_menu_overlay = False

        self.active_key_handler = self.sidebar
        self.key_handlers = []
        self.mouse_handlers = []
        self.last_mouse_event = ""

        self.render_group = [self.titlebar, self.sidebar, self.mainscreen, self.footerbar, self.title_menu_overlay,
                             self.debug_menu, self.help_menu, self.loader_status, self.file_browser, self.input_overlay]

        self.rebuild_all()

        self.stdscr.refresh()

    def ktool_dbg_print_func(self, msg):
        self.debug_menu.scroll_view_text_buffer.lines.append(msg)

    def ktool_dbg_print_err_func(self, msg):
        self.debug_menu.scroll_view_text_buffer.lines.append('§32m' + msg + '§39m')

    def setup(self):
        """
        Perform the curses initialization ritual
        :return: curses standard screen instance.
        """
        stdscr = curses.initscr()
        # Disable keypresses being printed on screen
        curses.noecho()
        # Normally, tty only passes the keypress buffer after a line break/etc. Disable that
        curses.cbreak()
        # Tell curses to give us all the funky keypress events too
        stdscr.keypad(True)
        # *And* all the funky mouse events
        curses.mousemask(curses.ALL_MOUSE_EVENTS)

        # Initialize color rendering stuff
        curses.start_color()
        # Don't break default terminal colors/transparency unless we need to.
        curses.use_default_colors()

        # hide the cursor
        curses.curs_set(0)

        self.supports_color = curses.has_colors()
        for i in range(0, curses.COLORS):
            try:
                curses.init_pair(i + 1, i, -1)
                self.supported_colors += 1
            except ValueError:
                pass

        return stdscr

    def teardown(self):
        """
        Perform the exact opposite of the setup() task. Return terminal to normalcy.

        IT IS ABSOLUTELY CRITICAL THAT THIS GETS CALLED ON EXIT. DO ABSOLUTELY EVERYTHING POSSIBLE TO TRY AND MAKE THAT
            HAPPEN. NOT DOING SO WILL REALLY FUCK UP TERM DISPLAY.
        :return:
        """
        curses.echo()
        curses.nocbreak()
        self.stdscr.keypad(False)
        curses.mousemask(0)
        curses.curs_set(1)
        curses.endwin()

    def update_mainscreen_text(self):
        """
        Pull lines from currently selected sidebar item and copy them over into the Main Screen.

        :return:
        """
        if len(self.sidebar.processed_items) < 1:
            return

        item = self.sidebar.processed_items[self.sidebar.selected_index]
        self.mainscreen.scroll_view_text_buffer.lines = item.content.lines

        if self.mainscreen.currently_displayed_index != self.sidebar.selected_index:
            self.mainscreen.currently_displayed_index = self.sidebar.selected_index
            self.mainscreen.scroll_view_text_buffer.scrollcursor = 0
            self.mainscreen.scroll_view_text_buffer.process_lines()
        self.mainscreen.set_tab_name(item.name)

    def update_load_status(self, msg):
        self.loader_status.status_string = msg
        self.redraw_all()

    def load_file(self, filename, mmap=True):
        """
        Load a file by filename into the GUI.

        :param filename:
        :return:
        """
        try:
            # KToolMachOLoader.SUPPORTS_256 = self.supported_colors > 200
            self.loader_status.draw = True
            self.mainscreen.scroll_view_text_buffer.lines = [f'Loading {filename}...']
            self.redraw_all()

            fd = open(filename, 'rb')

            self.mainscreen.set_tab_name(filename)

            filename_base = os.path.basename(filename)

            first1k = fd.read(0x1000)
            fd.seek(0)
            if b'__BOOTDATA\x00\x00\x00\x00\x00\x00' in first1k:
                for item in KToolKernelCacheLoader.contents_for_file(fd, self.update_load_status):
                    self.sidebar.add_menu_item(item)
            else:
                for item in KToolMachOLoader.contents_for_file(fd, self.update_load_status, mmap):
                    self.sidebar.add_menu_item(item)

            self.active_key_handler = self.sidebar
            self.key_handlers = [self.sidebar, self.mainscreen, self.titlebar, self.file_browser]
            self.mouse_handlers = [self.sidebar, self.titlebar, self.title_menu_overlay, self.debug_menu,
                                   self.input_overlay, self.help_menu]
            self.loader_status.draw = False
            self.input_overlay.draw = False

            self.redraw_all()

            self.program_loop()

        except Exception as ex:
            self.teardown()
            raise ex

    def rebuild_all(self):
        """
        Reconstruct all Views from the ground up (almost).

        This can be called on terminal resize to update the sizes of views/scroll buffers properly.

        Note: calls redraw_all() and refreshes the screen once finished.

        :return:
        """
        self.stdscr.clear()

        # curses doesn't update these values, so we do it manually here
        lines, cols = self.stdscr.getmaxyx()
        curses.LINES = lines
        curses.COLS = cols

        self.title_menu_overlay.box = Box(self.root, 0, 0, curses.COLS, curses.LINES)
        self.loader_status.box = Box(self.root, 0, 0, curses.COLS, curses.LINES)

        # Rebuild all of our contexts so they're drawn with updated screen width/height
        self.titlebar.box = Box(self.root, 0, 0, curses.COLS, 1)
        self.titlebar.exit_button.box.x = curses.COLS - 10
        self.titlebar.exit_button.box.parent = self.titlebar.box

        self.sidebar.box = Box(self.root, 0, 1, Sidebar.WIDTH, curses.LINES - 2)

        self.sidebar.scroll_view = Box(self.root, 1, 2, Sidebar.WIDTH - 2, curses.LINES - 4)
        self.sidebar.scroll_view_text_buffer = ScrollingDisplayBuffer(self.sidebar.scroll_view, 1, 0, Sidebar.WIDTH - 4,
                                                                      curses.LINES - 5)
        self.sidebar.scroll_view_text_buffer.wrap = False

        width = curses.COLS - Sidebar.WIDTH - 2
        self.mainscreen.box = Box(self.root, Sidebar.WIDTH, 1, width, curses.LINES - 2)
        self.mainscreen.info_box = Box(self.root, Sidebar.WIDTH + 1, 1, width - 1, 1)

        self.mainscreen.scroll_view = Box(self.root, Sidebar.WIDTH + 2, 3, width - 6, curses.LINES - 4)
        self.mainscreen.scroll_view_text_buffer = ScrollingDisplayBuffer(self.mainscreen.scroll_view, 1, 0, width - 8,
                                                                         curses.LINES - 5)

        self.debug_menu.box = Box(self.root, 5, 5, curses.COLS - 10, curses.LINES - 10)
        self.debug_menu.scroll_view = Box(self.root, 7, 6, curses.COLS - 12, curses.LINES - 12)

        ls = self.debug_menu.scroll_view_text_buffer.lines if self.debug_menu.scroll_view_text_buffer else []
        self.debug_menu.scroll_view_text_buffer = ScrollingDisplayBuffer(self.debug_menu.scroll_view, 0, 0,
                                                                         curses.COLS - 22, curses.LINES - 12)
        self.debug_menu.scroll_view_text_buffer.render_attr = curses.color_pair(1)
        self.debug_menu.scroll_view_text_buffer.lines = ls  # ?? We shouldn't need to do this
        self.debug_menu.parse_lines()

        self.help_menu.box = Box(self.root, 5, 5, curses.COLS - 10, curses.LINES - 10)
        self.help_menu.scroll_view = Box(self.root, 7, 6, curses.COLS - 12, curses.LINES - 12)

        self.help_menu.scroll_view_text_buffer = ScrollingDisplayBuffer(self.help_menu.scroll_view, 0, 0,
                                                                         curses.COLS - 22, curses.LINES - 12)
        self.help_menu.scroll_view_text_buffer.lines = MAIN_TEXT.split('\n')

        self.file_browser.box = Box(self.root, 0, 0, curses.COLS, curses.LINES)
        self.file_browser.scroll_view = Box(self.root, 5, 4, curses.COLS - 10, curses.LINES - 8)
        self.file_browser.scroll_view_text_buffer = ScrollingDisplayBuffer(self.file_browser.scroll_view, 0, 0,
                                                                           curses.COLS - 10, curses.LINES - 10)
        self.file_browser.scroll_view_text_buffer.wrap = False

        self.mainscreen.currently_displayed_index = -1
        self.update_mainscreen_text()

        self.footerbar.box = Box(self.root, 0, curses.LINES - 1, curses.COLS, 1)

        self.redraw_all()

    def redraw_all(self):
        """
        Wipe the screen and have the views redraw their contents

        :return:
        """

        self.stdscr.erase()

        self.update_mainscreen_text()

        self.footerbar.debug_text = f'{curses.COLS}x{curses.LINES} | {self.sidebar.selected_index} | {self.last_mouse_event} | self.titlebar.pres_menu_item = {str(self.titlebar.pres_menu_item)} '

        self.debug_menu.lines = [f'•Sself.titlebar.pres_menu_item = {str(self.titlebar.pres_menu_item)}']

        for item in self.render_group:
            if item.draw:
                item.redraw()

        self.stdscr.refresh()

    def handle_present_menu_exception(self, yes):
        if not yes:
            self.title_menu_overlay.draw = False
            self.is_showing_menu_overlay = False
            self.titlebar.pres_menu_item_index = -1
            self.active_key_handler = self.sidebar
        else:
            self.title_menu_overlay.draw = True
            self.title_menu_overlay.active_render_menu = self.titlebar.pres_menu_item[0]
            self.title_menu_overlay.active_menu_start_x = self.titlebar.pres_menu_item[1]
            self.is_showing_menu_overlay = True
            self.active_key_handler = self.titlebar

    def handle_mouse(self, x, y):
        self.last_mouse_event = f'M: x={x}, y={y}'
        for handler in self.mouse_handlers[::-1]:
            if handler.handle_mouse(x, y):
                break

    def handle_key_press(self, c):
        """
        Handle 'important' keys, pass the rest to current active subview

        :param c:
        :return:
        """

        if not self.active_key_handler.draw:
            self.active_key_handler = self.sidebar

        if c == curses.KEY_EXIT or c == curses.KEY_BACKSPACE:
            raise ExitProgramException

        elif c == curses.KEY_RESIZE:
            # Curses passes this weird keypress whenever the window gets resized
            # So, rebuild our contexts here with the new screen size.
            raise RebuildAllException

        if c == curses.KEY_MOUSE:
            _, mx, my, _, _ = curses.getmouse()
            self.handle_mouse(mx, my)

        elif c == ord('d'):
            if self.footerbar.show_debug:
                self.footerbar.show_debug = False
            else:
                self.footerbar.show_debug = True

        # TAB
        elif c == 9:
            if self.active_key_handler == self.sidebar:
                self.active_key_handler = self.mainscreen
                self.mainscreen.highlighted = True
            else:
                self.active_key_handler = self.sidebar
                self.mainscreen.highlighted = False

        else:
            if not self.active_key_handler.handle_key_press(c):
                for handler in self.key_handlers[::-1]:
                    if handler.handle_key_press(c):
                        break

    def program_loop(self):
        """
        Main Program Loop.

        1. Get a keypress ("keypress" includes mouse events because curses)
        2. Send the keypress to the handler, which will update object models, etc.
        2.5 Handle any exceptions raised by the keypress event.
        3. Redraw all of the views to update the contents of them.

        :return:
        """
        self.rebuild_all()

        while True:
            try:
                c = self.stdscr.getch()

                self.handle_key_press(c)

                self.redraw_all()

            except FileBrowserOpenNewFileException:
                self.file_browser.draw = True
                self.handle_present_menu_exception(False)
                self.active_key_handler = self.file_browser
                self.redraw_all()

            except RebuildAllException:
                self.rebuild_all()

            except PresentTitleMenuException:
                self.handle_present_menu_exception(True)
                try:
                    self.redraw_all()
                # this is so bad
                except HelpMenuException:
                    self.handle_present_menu_exception(False)
                    self.help_menu.draw = True
                    self.active_key_handler = self.help_menu
                    try:
                        self.redraw_all()
                    except PanicException:
                        self.teardown()
                        print(PANIC_STRING)
                        exit(1)

            except DestroyTitleMenuException:
                self.handle_present_menu_exception(False)
                self.redraw_all()

            except PresentDebugMenuException:
                self.debug_menu.draw = True
                self.active_key_handler = self.debug_menu
                try:
                    self.redraw_all()
                except PanicException:
                    self.teardown()
                    print(PANIC_STRING)
                    exit(1)

            except ExitProgramException:
                break

            except KeyboardInterrupt:
                break

            except PanicException:
                self.teardown()
                print(PANIC_STRING)
                exit(1)

            except Exception as ex:
                self.teardown()
                raise ex

        self.teardown()


def external_hard_fault_teardown():
    """
    Call this from outside of this file wherever the Screen is being loaded from, if it catches an Exception
        (it should never do that, so something went very wrong, and we need to attempt to unfuck the terminal).

    :return:
    """
    curses.echo()
    curses.nocbreak()
    curses.mousemask(0)
    curses.curs_set(1)
    curses.endwin()



#
#  ktool | ktool
#  kcache.py
#
#
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2022.
#
from io import BytesIO

import ktool
from kmacho.structs import *
from ktool import MachOFile, Image, log
from ktool.dyld import ImageHeader, Dyld
import ktool.kplistlib as plistlib
from ktool.exceptions import UnsupportedFiletypeException
from ktool.macho import SlicedBackingFile


class kmod_info_64(Struct):
    """
    """
    _FIELDNAMES = ['next_addr', 'info_version', 'id', 'name', 'version', 'reference_count', 'reference_list_addr',
                   'address', 'size', 'hdr_size', 'start_addr', 'stop_addr']
    _SIZES = [uint64_t, int32_t, uint32_t, char_t[64], char_t[64], int32_t, uint64_t,
              uint64_t, uint64_t, uint64_t, uint64_t, uint64_t]
    SIZE = sum([0xffff & i for i in _SIZES])

    def __init__(self, byte_order="little"):
        super().__init__(fields=self._FIELDNAMES, sizes=self._SIZES, byte_order=byte_order)


class Kext:
    def __init__(self):

        self.prelink_info = {}

        self.name = ""
        self.version = ""
        self.start_addr = 0

        self.development_region = ""
        self.executable_name = ""
        self.id = ""
        self.bundle_name = ""
        self.package_type = ""
        self.info_string = ""
        self.version_str = ""

        self.image = None


class EmbeddedKext(Kext):
    def __init__(self, image, prelink_info):
        super().__init__()
        self.start_addr = prelink_info['_PrelinkExecutableLoadAddr']
        self.size = prelink_info['_PrelinkExecutableSize']
        self.name = prelink_info['CFBundleIdentifier']
        self.version = prelink_info['CFBundleVersion']

        self.backing_file = BytesIO()
        self.backing_file.write(image.get_bytes_at(self.start_addr, self.size, vm=True))
        self.backing_file.seek(0)
        self.image = ktool.load_image(self.backing_file)


class MergedKext(Kext):
    def __init__(self, image: Image, kmod_info, start_addr):
        super().__init__()

        self.backing_image = image
        self.backing_slice = image.slice

        is64 = image.macho_header.is64
        self.name = image.get_cstr_at(kmod_info.off + (0x10 if is64 else 0x8), vm=False)
        self.version = image.get_cstr_at(kmod_info.off + 64 + (0x10 if is64 else 0x8), vm=False)
        self.start_addr = start_addr
        self.info = kmod_info

        file_base_addr = image.vm.translate(start_addr)

        # cool. we have a basic set of stuff in place, lets bootstrap up an Image from it.

        self.mach_header = ImageHeader.from_image(self.backing_slice, file_base_addr)
        self.image = Image(self.backing_slice)
        self.image.macho_header = self.mach_header
        self.image.vm_realign(yell_about_misalignment=False)

        # noinspection PyProtectedMember
        Dyld._parse_load_commands(self.image)
        # noinspection PyProtectedMember
        Dyld._process_image(self.image)

        for segment in image.segments.values():
            segment.vm_address = segment.vm_address | 0xffff000000000000


class KernelCache:

    def __init__(self, macho_file: MachOFile):
        self.mach_kernel_file = macho_file
        self.mach_kernel = ktool.load_image(macho_file)

        if self.mach_kernel.macho_header.is64:
            self.mach_kernel.vm.detag_kern_64 = True

        self.kexts = []

        self.prelink_info = {}

        if '__info' in self.mach_kernel.segments['__PRELINK_INFO'].sections:
            self._process_prelink_info()

        self.version = self.prelink_info['com.apple.kpi.mach']['CFBundleVersion']

        self.version_str = ""
        vloc = self.mach_kernel.slice.find('@(#)VERSION:')
        self.version_str = self.mach_kernel.get_cstr_at(vloc)
        dat = self.version_str.split('xnu_')[-1].split('/')[-1].lower()

        self.release_type = dat.split('_')[0]
        self.arch = dat.split('_')[1]
        self.soc = dat.split('_')[2]

        if '__kmod_info' in self.mach_kernel.segments['__PRELINK_INFO'].sections:
            self._process_merged_kexts()

        if len(self.kexts) == 0:
            if '_PrelinkExecutableLoadAddr' in self.prelink_info['com.apple.kpi.mach']:
                self._process_kexts_from_prelink_info()

        self._process_kexts()

    def _process_kexts_from_prelink_info(self):
        for kext_name, kext in self.prelink_info.items():
            try:
                self.kexts.append(EmbeddedKext(self.mach_kernel, kext))
            except UnsupportedFiletypeException:
                log.debug(f'Bad Header(?) at {kext_name}')
            except KeyError:
                pass

    def _process_kexts(self):
        for kext in self.kexts:
            if kext.name in self.prelink_info.keys():
                kext.executable_name = self.prelink_info[kext.name]['CFBundleExecutable']
                kext.id = self.prelink_info[kext.name]['CFBundleIdentifier']
                kext.bundle_name = self.prelink_info[kext.name]['CFBundleName']
                kext.package_type = self.prelink_info[kext.name]['CFBundlePackageType']
                kext.info_string = self.prelink_info[kext.name]['CFBundleGetInfoString'] if 'CFBundleGetInfoString' in self.prelink_info[kext.name] else ''
                kext.version_str = self.prelink_info[kext.name]['CFBundleVersion']

                kext.prelink_info = self.prelink_info[kext.name]

    def _process_prelink_info(self):
        address = self.mach_kernel.segments['__PRELINK_INFO'].sections['__info'].vm_address
        prelink_info_str = f'<plist version="1.0">{self.mach_kernel.get_cstr_at(address, vm=True)}</plist>'
        prelink_info_dat = prelink_info_str.encode('utf-8')
        prelink_info = plistlib.readPlistFromBytes(prelink_info_dat)
        items = prelink_info['_PrelinkInfoDictionary']
        for bundle_dict in items:
            self.prelink_info[bundle_dict['CFBundleIdentifier']] = bundle_dict

    def _process_merged_kexts(self):
        kext_starts = []
        kmod_start_sect = self.mach_kernel.segments['__PRELINK_INFO'].sections['__kmod_start']

        ptr_size = 8 if self.mach_kernel.macho_header.is64 else 4

        for i in range(kmod_start_sect.file_address, kmod_start_sect.file_address + kmod_start_sect.size, ptr_size):
            kext_starts.append(self.mach_kernel.get_int_at(i, ptr_size, vm=False))

        kmod_info_locations = []
        kmod_info_sect = self.mach_kernel.segments['__PRELINK_INFO'].sections['__kmod_info']

        for i in range(kmod_info_sect.file_address, kmod_info_sect.file_address + kmod_info_sect.size, ptr_size):
            kmod_info_locations.append(self.mach_kernel.get_int_at(i, ptr_size, vm=False))

        # start processing kmod info
        for i, info_loc in enumerate(kmod_info_locations):
            info = self.mach_kernel.load_struct(info_loc, kmod_info_64, vm=True)

            start_addr = kext_starts[i]
            kext = MergedKext(self.mach_kernel, info, start_addr)
            self.kexts.append(kext)



#
#  ktool | ktool
#  dyld.py
#
#  This file includes a lot of utilities, classes, and abstractions
#  designed for replicating certain functionality within dyld.
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2021.
#
import math
from collections import namedtuple
from typing import List, Union, Dict

from kmacho import (
    MH_FLAGS,
    MH_FILETYPE,
    LOAD_COMMAND,
    BINDING_OPCODE,
    LOAD_COMMAND_MAP,
    BIND_SUBOPCODE_THREADED_SET_BIND_ORDINAL_TABLE_SIZE_ULEB,
    BIND_SUBOPCODE_THREADED_APPLY, MH_MAGIC_64, CPUType, CPUSubTypeARM64
)
from kmacho.base import Constructable
from kmacho.fixups import *
from ktool.codesign import CodesignInfo
from ktool.macho import Segment, Slice, VM, MisalignedVM
from ktool.util import log, macho_is_malformed, ignore, bytes_to_hex


class ImageHeader(Constructable):
    """
    This class represents the Mach-O Header
    It contains the basic header info along with all load commands within it.

    It doesn't handle complex abstraction logic, it simply loads in the load commands as their raw structs
    """

    @classmethod
    def from_image(cls, macho_slice, offset=0) -> 'ImageHeader':

        image_header = ImageHeader()

        header: mach_header = macho_slice.load_struct(offset, mach_header)

        if header.magic == MH_MAGIC_64:
            header: mach_header_64 = macho_slice.load_struct(offset, mach_header_64)
            image_header.is64 = True

        raw = header.raw

        image_header.filetype = MH_FILETYPE(header.filetype)

        for flag in MH_FLAGS:
            if header.flags & flag.value:
                image_header.flags.append(flag)

        offset += header.SIZE

        load_commands = []

        for i in range(1, header.loadcnt + 1):
            cmd = macho_slice.get_int_at(offset, 4)
            cmd_size = macho_slice.get_int_at(offset + 4, 4)

            cmd_raw = macho_slice.get_bytes_at(offset, cmd_size)
            try:
                load_cmd = Struct.create_with_bytes(LOAD_COMMAND_MAP[LOAD_COMMAND(cmd)], cmd_raw)
                load_cmd.off = offset
            except ValueError as ex:
                if not ignore.MALFORMED:
                    raise ex
                unk_lc = macho_slice.load_struct(offset, unk_command)
                load_cmd = unk_lc
            except KeyError as ex:
                if not ignore.MALFORMED:
                    log.error()
                    log.error(f'Load Command {str(LOAD_COMMAND(cmd))} doesn\'t have a mapped struct type')
                    log.error('*Please* file an issue on the github @ https://github.com/cxnder/ktool')
                    log.error()
                    log.error(f'Run with the -f flag before the subcommand to try and force loading anyways')
                    log.error()
                    raise ex
                unk_lc = macho_slice.load_struct(offset, unk_command)
                load_cmd = unk_lc

            load_commands.append(load_cmd)
            raw += cmd_raw
            offset += load_cmd.cmdsize

        image_header.raw = raw
        image_header.dyld_header = header
        image_header.load_commands = load_commands

        return image_header

    @classmethod
    def from_values(cls, *args, **kwargs):
        pass

    def __str__(self):
        return f'MachO Header - 64 bit VM: {self.is64} | File Type: {self.filetype} | Flags: {self.flags} | Load Cmd Count: {len(self.load_commands)}'

    def __init__(self):
        self.is64 = False
        self.dyld_header = None
        self.filetype = MH_FILETYPE(0)
        self.flags: List[MH_FLAGS] = []
        self.load_commands = []
        self.raw = bytearray()

    def serialize(self):
        return {
            'filetype': self.filetype.name,
            'flags': [flag.name for flag in self.flags],
            'is_64_bit': self.is64,
            'dyld_header': self.dyld_header.serialize(),
            'load_commands': [cmd.serialize() for cmd in self.load_commands]
        }

    def raw_bytes(self) -> bytes:
        return self.raw


class Image:
    """
    This class represents the Mach-O Binary as a whole.

    It's the root object in the massive tree of information we're going to build up about the binary

    This class on its own does not handle populating its fields.
    The Dyld class set is responsible for loading in and processing the raw values to it.
    """

    def __init__(self, macho_slice: Slice):
        """
        Create a MachO image

        :param macho_slice: MachO Slice being processed
        :type macho_slice: MachO Slice
        """
        self.slice: Slice = macho_slice

        self.vm = None

        self.fallback_vm = MisalignedVM(self.slice)

        if self.slice:
            self.macho_header: ImageHeader = ImageHeader.from_image(macho_slice=macho_slice)

            self.vm_realign()

        self.base_name = ""  # copy of self.name
        self.install_name = ""

        self.linked_images: List[ExternalDylib] = []

        self.segments: Dict[str, Segment] = {}

        self.info: Union[dyld_info_command, None] = None
        self.dylib: Union[ExternalDylib, None] = None
        self.uuid = None
        self.codesign_info: Union[CodesignInfo, None] = None

        self._codesign_cmd = None

        self.platform: PlatformType = PlatformType.UNK

        self.allowed_clients: List[str] = []

        self.rpath: Union[str, None] = None

        self.minos = os_version(0, 0, 0)
        self.sdk_version = os_version(0, 0, 0)

        self.imports: List[Symbol] = []
        self.exports: List[Symbol] = []

        self.symbols: Dict[int, Symbol] = {}
        self.import_table: Dict[int, Symbol] = {}
        self.export_table: Dict[int, Symbol] = {}

        self.entry_point = 0

        self.function_starts: List[int] = []

        self.thread_state: List[int] = []
        self._entry_off = 0

        self.binding_table = None
        self.weak_binding_table = None
        self.lazy_binding_table = None
        self.export_trie: Union[ExportTrie, None] = None

        self.chained_fixups = None

        self.symbol_table: Union[SymbolTable, None] = None

        self.struct_cache: Dict[int, Struct] = {}

    def serialize(self):
        image_dict = {
            'macho_header': self.macho_header.serialize()
        }

        if self.install_name != "":
            image_dict['install_name'] = self.install_name

        linked = []
        for ext_dylib in self.linked_images:
            linked.append(ext_dylib.serialize())

        image_dict['linked'] = linked

        segments = {}

        for seg_name, seg in self.segments.items():
            segments[seg_name] = seg.serialize()

        image_dict['segments'] = segments
        if self.uuid:
            image_dict['uuid'] = bytes_to_hex(self.uuid)

        image_dict['platform'] = self.platform.name

        image_dict['allowed-clients'] = self.allowed_clients

        if self.rpath:
            image_dict['rpath'] = self.rpath

        image_dict['imports'] = [sym.serialize() for sym in self.imports]
        image_dict['exports'] = [sym.serialize() for sym in self.exports]
        image_dict['symbols'] = [sym.serialize() for sym in self.symbols.values()]

        image_dict['entry_point'] = self.entry_point

        image_dict['function_starts'] = self.function_starts

        image_dict['thread_state'] = self.thread_state

        image_dict['minos'] = f'{self.minos.x}.{self.minos.y}{self.minos.z}'
        image_dict['sdk_version'] = f'{self.sdk_version.x}.{self.sdk_version.y}.{self.sdk_version.z}'

        return image_dict

    def vm_realign(self, yell_about_misalignment=True):

        align_by = 0x4000
        aligned = False

        detag_64 = False

        segs = []
        for cmd in self.macho_header.load_commands:
            if cmd.cmd in [LOAD_COMMAND.SEGMENT.value, LOAD_COMMAND.SEGMENT_64.value]:
                segs.append(cmd)
            if cmd.cmd == LOAD_COMMAND.LC_DYLD_CHAINED_FIXUPS:
                detag_64 = True

        if self.slice.type == CPUType.ARM64 and self.slice.subtype == CPUSubTypeARM64.ARM64E:
            detag_64 = True

        while not aligned:
            aligned = True
            for cmd in segs:
                cmd: segment_command_64 = cmd
                if cmd.vmaddr % align_by != 0:
                    if align_by == 0x4000:
                        align_by = 0x1000
                        aligned = False
                        break
                    else:
                        align_by = 0
                        aligned = True
                        break

        if align_by != 0:
            log.info(f'Aligned to {hex(align_by)} pages')
            self.vm: VM = VM(page_size=align_by)
            self.vm.detag_64 = detag_64
            self.vm.fallback = self.fallback_vm
        else:
            if yell_about_misalignment:
                log.info("MachO cannot be aligned to 16k or 4k pages. Swapping to fallback mapping.")
            self.vm: MisalignedVM = self.fallback_vm
            self.vm.detag_64 = detag_64

    def vm_check(self, address):
        return self.vm.vm_check(address)

    def get_int_at(self, offset: int, length: int, vm=False, section_name=None):
        """
        Get a sequence of bytes (as an int) from a location

        :param offset: Offset within the image
        :param length: Amount of bytes to get
        :param vm: Is `offset` a VM address
        :param section_name: Section Name if vm==True (improves translation time slightly)
        :return: `length` Bytes at `offset`
        """
        if vm:
            offset = self.vm.translate(offset)
        return self.slice.get_int_at(offset, length)

    def get_bytes_at(self, offset: int, length: int, vm=False, section_name=None):
        """
        Get a sequence of bytes from a location

        :param offset: Offset within the image
        :param length: Amount of bytes to get
        :param vm: Is `offset` a VM address
        :param section_name: Section Name if vm==True (improves translation time slightly)
        :return: `length` Bytes at `offset`
        """
        if vm:
            offset = self.vm.translate(offset)
        return self.slice.get_bytes_at(offset, length)

    def load_struct(self, address: int, struct_type, vm=False, section_name=None, endian="little", force_reload=False):
        """
        Load a struct (struct_type_t) from a location and return the processed object

        :param address: Address to load struct from
        :param struct_type: type of struct (e.g. dyld_header)
        :param vm:  Is `address` a VM address?
        :param section_name: if `vm==True`, the section name (slightly improves translation speed)
        :param endian: Endianness of bytes to read.
        :return: Loaded struct
        """
        if address not in self.struct_cache or force_reload:
            if vm:
                address = self.vm.translate(address)
            struct = self.slice.load_struct(address, struct_type, endian)
            self.struct_cache[address] = struct
            return struct

        return self.struct_cache[address]

    def get_str_at(self, address: int, count: int, vm=False, section_name=None, force=False):
        """
        Get string with set length from location (to be used essentially only for loading segment names)

        :param address: Address of string start
        :param count: Length of string
        :param vm: Is `address` a VM address?
        :param section_name: if `vm==True`, the section name (unused here, really)
        :return: The loaded string.
        """
        if vm:
            address = self.vm.translate(address)
        return self.slice.get_str_at(address, count, force=force)

    def get_cstr_at(self, address: int, limit: int = 0, vm=False, section_name=None):
        """
        Load a C style string from a location, stopping once a null byte is encountered.

        :param address: Address to load string from
        :param limit: Limit of the length of bytes, 0 = unlimited
        :param vm: Is `address` a VM address?
        :param section_name: if `vm==True`, the section name (vastly improves VM lookup time)
        :return: The loaded C string
        """
        if vm:
            address = self.vm.translate(address)
        return self.slice.get_cstr_at(address, limit)

    def decode_uleb128(self, readHead: int):
        """
        Decode a uleb128 integer from a location

        :param readHead: Start location
        :return: (end location, value)
        """
        return self.slice.decode_uleb128(readHead)


class Dyld:
    """
    This class takes our initialized "Image" object, parses through the raw data behind it, and fills out its properties.

    """

    @classmethod
    def load(cls, macho_slice: Slice, load_symtab=True, load_imports=True, load_exports=True) -> Image:
        """
        Take a slice of a macho file and process it using the dyld functions

        :param load_exports: Load Exports
        :param load_imports: Load Imports
        :param load_symtab: Load Symbol Table
        :param macho_slice: Slice to load. If your image is not fat, that'll be MachOFile.slices[0]
        :type macho_slice: Slice
        :return: Processed image object
        :rtype: Image
        """
        log.info("Loading image")
        image = Image(macho_slice)

        log.info("Processing Load Commands")
        Dyld._parse_load_commands(image, load_symtab, load_imports, load_exports)

        log.info("Processing Image")
        Dyld._process_image(image)
        return image

    @classmethod
    def _parse_load_commands(cls, image: Image, load_symtab=True, load_imports=True, load_exports=True) -> None:
        # noinspection PyUnusedLocal
        fixups = None
        log.info(f'registered {len(image.macho_header.load_commands)} Load Commands')
        for cmd in image.macho_header.load_commands:
            try:
                load_command = LOAD_COMMAND(cmd.cmd)
            except ValueError:
                continue

            if load_command == LOAD_COMMAND.SEGMENT_64 or load_command == LOAD_COMMAND.SEGMENT:
                log.debug_tm("Loading Segment")
                segment = Segment(image, cmd)

                log.info(f'Loaded Segment {segment.name}')
                image.vm.add_segment(segment)
                image.segments[segment.name] = segment

            elif load_command in [LOAD_COMMAND.THREAD, LOAD_COMMAND.UNIXTHREAD]:
                thread_state = []
                for i in range(cmd.count):
                    off = cmd.off + 16 + (i * 4)
                    val = image.get_int_at(off, 4)
                    thread_state.append(val)

                image.thread_state = thread_state

            elif load_command == LOAD_COMMAND.CODE_SIGNATURE:
                image._codesign_cmd = cmd
                image.codesign_info = CodesignInfo.from_image(image, cmd)

            elif load_command == LOAD_COMMAND.MAIN:
                image._entry_off = cmd.entryoff

            elif load_command == LOAD_COMMAND.DYLD_INFO_ONLY:
                image.info = cmd

                if load_imports:
                    log.info("Loading Binding Info")
                    image.binding_table = BindingTable(image, cmd.bind_off, cmd.bind_size)
                    image.weak_binding_table = BindingTable(image, cmd.weak_bind_off, cmd.weak_bind_size)
                    image.lazy_binding_table = BindingTable(image, cmd.lazy_bind_off, cmd.lazy_bind_size)

                if load_exports:
                    log.info("Loading Export Trie")
                    image.export_trie = ExportTrie.from_image(image, cmd.export_off, cmd.export_size)

            elif load_command == LOAD_COMMAND.FUNCTION_STARTS:
                fs_start = cmd.dataoff
                fs_size = cmd.datasize
                read_head = fs_start

                fs_addr = image.vm.vm_base_addr

                while read_head < fs_start + fs_size:
                    fs_r_addr, read_head = image.decode_uleb128(read_head)
                    fs_addr += fs_r_addr
                    image.function_starts.append(fs_addr)

            elif load_command == LOAD_COMMAND.LC_DYLD_EXPORTS_TRIE:
                log.info("Loading Export Trie")
                image.export_trie = ExportTrie.from_image(image, cmd.dataoff, cmd.datasize)

            elif load_command == LOAD_COMMAND.LC_DYLD_CHAINED_FIXUPS:
                if load_imports:
                    image.chained_fixups = ChainedFixups.from_image(image, cmd)

            elif load_command == LOAD_COMMAND.SYMTAB:
                if load_symtab:
                    log.info("Loading Symbol Table")
                    image.symbol_table = SymbolTable(image, cmd)

            elif load_command == LOAD_COMMAND.DYSYMTAB:
                pass

            elif load_command == LOAD_COMMAND.UUID:
                image.uuid = cmd.uuid
                log.info(f'image UUID: {image.uuid}')

            elif load_command == LOAD_COMMAND.SUB_CLIENT:
                string = image.get_cstr_at(cmd.off + cmd.offset)
                image.allowed_clients.append(string)
                log.debug(f'Loaded Subclient "{string}"')

            elif load_command == LOAD_COMMAND.RPATH:
                string = image.get_cstr_at(cmd.off + cmd.path)
                image.rpath = string
                log.info(f'image Resource Path: {string}')

            elif load_command == LOAD_COMMAND.BUILD_VERSION:
                image.platform = PlatformType(cmd.platform)
                image.minos = os_version(x=image.get_int_at(cmd.off + 14, 2), y=image.get_int_at(cmd.off + 13, 1),
                                         z=image.get_int_at(cmd.off + 12, 1))
                image.sdk_version = os_version(x=image.get_int_at(cmd.off + 18, 2),
                                               y=image.get_int_at(cmd.off + 17, 1),
                                               z=image.get_int_at(cmd.off + 16, 1))
                log.info(f'Loaded platform {image.platform.name} | '
                         f'Minimum OS {image.minos.x}.{image.minos.y}'
                         f'.{image.minos.z} | SDK Version {image.sdk_version.x}'
                         f'.{image.sdk_version.y}.{image.sdk_version.z}')

            elif isinstance(cmd, version_min_command):
                # Only override this if it wasn't set by build_version
                if image.platform == PlatformType.UNK:
                    if load_command == LOAD_COMMAND.VERSION_MIN_MACOSX:
                        image.platform = PlatformType.MACOS
                    elif load_command == LOAD_COMMAND.VERSION_MIN_IPHONEOS:
                        image.platform = PlatformType.IOS
                    elif load_command == LOAD_COMMAND.VERSION_MIN_TVOS:
                        image.platform = PlatformType.TVOS
                    elif load_command == LOAD_COMMAND.VERSION_MIN_WATCHOS:
                        image.platform = PlatformType.WATCHOS

                    image.minos = os_version(x=image.get_int_at(cmd.off + 10, 2),
                                             y=image.get_int_at(cmd.off + 9, 1),
                                             z=image.get_int_at(cmd.off + 8, 1))

            elif load_command == LOAD_COMMAND.ID_DYLIB:
                image.dylib = ExternalDylib(image, cmd)
                log.debug(f'Loaded local dylib_command with install_name {image.dylib.install_name}')

            elif isinstance(cmd, dylib_command):
                # noinspection PyTypeChecker
                external_dylib = ExternalDylib(image, cmd)

                image.linked_images.append(external_dylib)
                log.debug(f'Loaded linked dylib_command with install name {external_dylib.install_name}')

    @staticmethod
    def _process_image(image: Image) -> None:
        """
        Once all load commands have been processed, process the results.
        This is mainly for things which need to be done once *all* lcs have been processed.

        :param image:
        :return:
        """
        if image.dylib is not None:
            image.name = image.dylib.install_name.split('/')[-1]
            image.base_name = image.dylib.install_name.split('/')[-1]
            image.install_name = image.dylib.install_name
        else:
            image.name = ""
            image.base_name = image.slice.file.name
            image.install_name = ""

        if image.export_trie:
            for symbol in image.export_trie.symbols:
                image.exports.append(symbol)
                image.export_table[symbol.address] = symbol

        if image.binding_table:
            for symbol in image.binding_table.symbol_table:
                symbol.attr = ''
                image.imports.append(symbol)
                image.import_table[symbol.address] = symbol
            for symbol in image.weak_binding_table.symbol_table:
                symbol.attr = 'Weak'
                image.imports.append(symbol)
                image.import_table[symbol.address] = symbol
            for symbol in image.lazy_binding_table.symbol_table:
                symbol.attr = 'Lazy'
                image.imports.append(symbol)
                image.import_table[symbol.address] = symbol

        if image.chained_fixups:
            for symbol in image.chained_fixups.symbols:
                symbol.attr = ''
                image.imports.append(symbol)
                image.import_table[symbol.address] = symbol

        if image.symbol_table:
            for symbol in image.symbol_table.table:
                image.symbols[symbol.address] = symbol

        # noinspection PyProtectedMember
        if len(image.thread_state) > 0:
            image.entry_point = image.thread_state[-4] if image.macho_header.is64 else image.thread_state[-2]

        elif image._entry_off > 0:
            # noinspection PyProtectedMember
            image.entry_point = image.vm.vm_base_addr + image._entry_off


class LD64:
    @classmethod
    def insert_load_cmd(cls, image: Image, lc, fields, index=-1):
        lc_type = LOAD_COMMAND_MAP[lc]

        load_cmd = Struct.create_with_values(lc_type, [lc.value, lc_type.SIZE] + fields)

        off = mach_header.SIZE
        off += image.macho_header.dyld_header.loadsize
        raw = load_cmd.raw
        size = len(load_cmd.raw)

        if index != -1:
            b_load_cmd = image.macho_header.load_commands[index - 1]
            off = b_load_cmd.off + b_load_cmd.cmdsize
            after_bytes = image.macho_header.raw_bytes()[off:image.macho_header.dyld_header.loadsize + 32]
            image.slice.patch(off, raw)
            image.slice.patch(off + size, after_bytes)
        else:
            image.slice.patch(off, raw)

        image.macho_header.load_commands.append(load_cmd)

        image.macho_header.dyld_header.loadcnt += 1
        image.macho_header.dyld_header.loadsize += size

        image.slice.patch(image.macho_header.dyld_header.off, image.macho_header.dyld_header.raw[:image.macho_header.dyld_header.__class__.SIZE])

    @classmethod
    def insert_load_cmd_with_str(cls, image: Image, lc, fields, suffix, index=-1):
        lc_type = LOAD_COMMAND_MAP[lc]
        load_cmd = Struct.create_with_values(lc_type, [lc.value, lc_type.SIZE] + fields)

        encoded = suffix.encode('utf-8') + b'\x00'
        cmd_size = 0x8 * math.ceil(lc_type.SIZE + len(encoded) / 0x8)

        header_size = image.segments['__TEXT'].sections['__text'].file_address

        # verify our lc will actually fit
        assert image.macho_header.dyld_header.loadsize + len(image.macho_header.dyld_header.raw)  + cmd_size < header_size

        load_cmd.cmdsize = cmd_size
        raw = load_cmd.raw + encoded + (b'\x00' * (cmd_size - (lc_type.SIZE + len(encoded))))

        full_header = bytearray()

        pre_load_cmd = image.macho_header.load_commands[index - 1] if index != -1 else image.macho_header.load_commands[-1]
        off = pre_load_cmd.off + pre_load_cmd.cmdsize

        full_header += bytearray(image.macho_header.raw_bytes()[:off])
        patch_after = bytearray()

        if index != -1:
            patch_after = bytearray(image.macho_header.raw_bytes()[off:image.macho_header.dyld_header.loadsize + len(image.macho_header.dyld_header.raw)])

        full_header += bytearray(raw)
        full_header += patch_after

        image.macho_header.dyld_header.loadcnt += 1
        image.macho_header.dyld_header.loadsize -= len(raw)

        full_header = bytearray(image.macho_header.dyld_header.raw) + full_header[len(image.macho_header.dyld_header.raw):]

        image.slice.patch(0, bytes(full_header))

    @classmethod
    def remove_load_command(cls, image: Image, index):
        b_load_cmd = image.macho_header.load_commands.pop(index)

        full_header = bytearray()

        off = b_load_cmd.off + b_load_cmd.cmdsize
        full_header += bytearray(image.macho_header.raw_bytes()[:b_load_cmd.off])  # add everything up to the target
        full_header += bytearray(image.macho_header.raw_bytes()[off:])  # add everything after the target command
        full_header += bytearray(b'\x00' * b_load_cmd.cmdsize)  # replace the now junk copies of displaced commands

        image.macho_header.dyld_header.loadcnt -= 1
        image.macho_header.dyld_header.loadsize -= b_load_cmd.cmdsize

        full_header = bytearray(image.macho_header.dyld_header.raw) + full_header[len(image.macho_header.dyld_header.raw):]

        image.slice.patch(0, bytes(full_header))


class ExternalDylib:
    def __init__(self, source_image: Image, cmd):
        self.cmd = cmd
        self.source_image = source_image

        self.install_name = self._get_name(cmd)
        self.weak = cmd.cmd == LOAD_COMMAND.LOAD_WEAK_DYLIB.value
        self.local = cmd.cmd == LOAD_COMMAND.ID_DYLIB.value

    def serialize(self):
        return {
            'install_name': self.install_name,
            'load_command': LOAD_COMMAND(self.cmd.cmd).name
        }

    def _get_name(self, cmd) -> str:
        read_address = cmd.off + dylib_command.SIZE
        return self.source_image.get_cstr_at(read_address)


os_version = namedtuple("os_version", ["x", "y", "z"])


class PlatformType(Enum):
    MACOS = 1
    IOS = 2
    TVOS = 3
    WATCHOS = 4
    BRIDGE_OS = 5
    MAC_CATALYST = 6
    IOS_SIMULATOR = 7
    TVOS_SIMULATOR = 8
    WATCHOS_SIMULATOR = 9
    DRIVER_KIT = 10
    UNK = 64


class ToolType(Enum):
    CLANG = 1
    SWIFT = 2
    LD = 3


class SymbolType(Enum):
    CLASS = 0
    METACLASS = 1
    IVAR = 2
    FUNC = 3
    UNK = 4


class Symbol(Constructable):
    """
    This class can represent several types of symbols.

    """

    @classmethod
    def from_image(cls, image, cmd, entry):
        fullname = image.get_cstr_at(entry.str_index + cmd.stroff)
        addr = entry.value

        symbol = cls.from_values(fullname, addr)

        N_STAB = 0xe0
        N_PEXT = 0x10
        N_TYPE = 0x0e
        N_EXT = 0x01

        type_masked = N_TYPE & entry.type
        for name, flag in {'N_UNDF': 0x0, 'N_ABS': 0x2, 'N_SECT': 0xe, 'N_PBUD': 0xc, 'N_INDR': 0xa}.items():
            if type_masked & flag:
                symbol.types.append(name)

        if entry.type & N_EXT:
            symbol.external = True

        return symbol

    @classmethod
    def from_values(cls, fullname, value, external=False, ordinal=0):
        if '_$_' in fullname:
            if fullname.startswith('_OBJC_CLASS_$'):
                dec_type = SymbolType.CLASS
            elif fullname.startswith('_OBJC_METACLASS_$'):
                dec_type = SymbolType.METACLASS
            elif fullname.startswith('_OBJC_IVAR_$'):
                dec_type = SymbolType.IVAR
            else:
                dec_type = SymbolType.UNK
            name = fullname.split('$')[1]
        else:
            name = fullname
            dec_type = SymbolType.FUNC

        return cls(fullname, name=name, dec_type=dec_type, external=external, value=value, ordinal=ordinal)

    def raw_bytes(self):
        pass

    def serialize(self):
        return {
            'name': self.fullname,
            'address': self.address,
            'external': self.external,
            'ordinal': self.ordinal
        }

    def __init__(self, fullname=None, name=None, dec_type=None, external=False, value=0, ordinal=0):
        self.fullname = fullname
        self.name = name
        self.dec_type = dec_type

        self.address = value

        self.entry = None
        self.ordinal = ordinal

        self.types = []
        self.external = external

        self.attr = None


class SymbolTable:
    """
    This class represents the symbol table declared in the MachO File

    .table contains the symbol table

    .ext contains exported symbols, i think?

    This class is incomplete

    """

    def __init__(self, image: Image, cmd: symtab_command):
        self.image: Image = image
        self.cmd: symtab_command = cmd

        self.ext: List[Symbol] = []
        self.table: List[Symbol] = self._load_symbol_table()

    def _load_symbol_table(self) -> List[Symbol]:
        symbol_table = []
        read_address = self.cmd.symoff
        typing = symtab_entry if self.image.macho_header.is64 else symtab_entry_32

        for i in range(0, self.cmd.nsyms):
            entry = self.image.load_struct(read_address + typing.SIZE * i, typing)
            symbol = Symbol.from_image(self.image, self.cmd, entry)
            symbol_table.append(symbol)

            if symbol.external:
                self.ext.append(symbol)

            log.debug_tm(f'Symbol Table: Loaded symbol:{symbol.name} ordinal:{symbol.ordinal} type:{symbol.dec_type}')
            log.debug_tm(str(entry))

        return symbol_table


class ChainedFixups(Constructable):
    @classmethod
    def from_image(cls, image: Image, chained_fixup_cmd: linkedit_data_command):

        symbols = []

        import struct

        fixup_hdr_addr = chained_fixup_cmd.dataoff
        fixup_hdr = image.load_struct(fixup_hdr_addr, dyld_chained_fixups_header, vm=False)

        imports_format = dyld_chained_import_format(fixup_hdr.imports_format)

        segs_addr = fixup_hdr_addr + fixup_hdr.starts_offset
        seg_count = image.get_int_at(segs_addr, 4, vm=False)
        imports_addr = fixup_hdr_addr + fixup_hdr.imports_offset

        syms_addr = fixup_hdr_addr + fixup_hdr.symbols_offset

        segs = []
        for i in range(seg_count):
            s = image.get_int_at((i * 4) + segs_addr + 4, 4)  # read
            segs.append(s)

        for i in range(seg_count):
            if segs[i] == 0:
                continue
            starts_addr = (
                    segs_addr + segs[i]
            )

            starts_in_segment_data = image.get_bytes_at(starts_addr, 24)

            starts_in_segment = struct.unpack("<IHHQIHH", starts_in_segment_data)

            page_count = starts_in_segment[5]
            page_size = starts_in_segment[1]
            segment_offset = starts_in_segment[3]
            pointer_type = starts_in_segment[2]

            PTR_STRUCT_TYPE_BASE = DYLD_CHAINED_PTR_BASE[dyld_chained_ptr_format(pointer_type)]
            PTR_STRUCT_TYPE_FUNC = DYLD_CHAINED_PTR_FMATS[dyld_chained_ptr_format(pointer_type)]

            page_starts_data = image.get_bytes_at(starts_addr + 22, page_count * 2)
            page_starts = struct.unpack("<" + ("H" * page_count), page_starts_data)

            for (j, start) in enumerate(page_starts):
                if start == DYLD_CHAINED_PTR_START_NONE:
                    continue

                chain_entry_addr = (
                        segment_offset + (j * page_size) + start
                )

                j += 1
                while True:
                    content = image.get_int_at(chain_entry_addr, 8)
                    item = image.load_struct(chain_entry_addr, PTR_STRUCT_TYPE_BASE, vm=False)
                    item = image.load_struct(chain_entry_addr, PTR_STRUCT_TYPE_FUNC(item), vm=False, force_reload=True)

                    offset = content & 0xFFFFFFFF
                    nxt = (content >> 50) & 2047
                    bind = (content >> 62) & 1
                    if bind == 1:
                        import_entry = image.get_int_at(imports_addr + offset * 4, 4)
                        ordinal = import_entry & 0xFF
                        sym_name_offset = import_entry >> 9
                        sym_name_addr = syms_addr + sym_name_offset

                        sym_name = image.get_cstr_at(
                            sym_name_addr)
                        sym = Symbol.from_values(sym_name, sym_name_addr, True, ordinal)
                        symbols.append(sym)

                    else:
                        pass

                    # next tells us how many u32 until the next chain entry
                    skip = nxt * 4
                    chain_entry_addr += skip
                    # if skip == 0, chain is done
                    if skip == 0:
                        break

        return cls(symbols)

    @classmethod
    def from_values(cls, *args, **kwargs):
        pass

    def raw_bytes(self):
        pass

    def __init__(self, symbols):
        self.symbols = symbols


export_node = namedtuple("export_node", ['text', 'offset', 'flags'])


class ExportTrie(Constructable):
    @classmethod
    def from_image(cls, image: Image, export_start: int, export_size: int) -> 'ExportTrie':
        trie = ExportTrie()

        endpoint = export_start + export_size
        nodes = ExportTrie.read_node(image, export_start, '', export_start, endpoint)
        symbols = []

        for node in nodes:
            if node.text:
                symbols.append(Symbol.from_values(node.text, node.offset, False))

        trie.nodes = nodes
        trie.symbols = symbols
        trie.raw = image.get_bytes_at(export_start, export_size)

        return trie

    @classmethod
    def from_values(cls, *args, **kwargs):
        pass

    def raw_bytes(self):
        return self.raw

    def __init__(self):
        self.raw = bytearray()
        self.nodes: List[export_node] = []
        self.symbols: List[Symbol] = []

    @classmethod
    def read_node(cls, image: Image, trie_start: int, string: str, cursor: int, endpoint: int) -> List[export_node]:

        if cursor > endpoint:
            log.error("Node offset greater than size of export trie")
            macho_is_malformed()

        start = cursor
        byte = image.get_int_at(cursor, 1)
        results = []
        log.debug_tm(f'@ {hex(start)} node: {hex(byte)} current_symbol: {string}')
        if byte == 0:
            cursor += 1
            branches = image.get_int_at(cursor, 1)
            log.debug_tm(f'BRAN {branches}')
            for i in range(0, branches):
                if i == 0:
                    cursor += 1
                proc_str = image.get_cstr_at(cursor)
                cursor += len(proc_str) + 1
                offset, cursor = image.decode_uleb128(cursor)
                log.debug_tm(f'({i}) string: {string + proc_str} next_node: {hex(trie_start + offset)}')
                results += ExportTrie.read_node(image, trie_start, string + proc_str, trie_start + offset, endpoint)
        else:
            log.debug_tm(f'TERM: 0')
            size, cursor = image.decode_uleb128(cursor)
            flags = image.get_int_at(cursor, 1)
            cursor += 1
            offset, cursor = image.decode_uleb128(cursor)
            results.append(export_node(string, offset, flags))

        return results


action = namedtuple("action", ["vmaddr", "libname", "item"])
record = namedtuple("record", [
    "off",
    "seg_index",
    "seg_offset",
    "lib_ordinal",
    "type",
    "flags",
    "name",
    "addend",
    "special_dylib"
])


class BindingTable:
    """
    The binding table contains a ton of information related to the binding info in the image

    .lookup_table - Contains a map of address -> Symbol declarations which should be used for processing off-image
    symbol decorations

    .symbol_table - Contains a full list of symbols declared in the binding info. Avoid iterating through this for
    speed purposes.

    .actions - contains a list of, you guessed it, actions.

    .import_stack - contains a fairly raw unprocessed list of binding info commands

    """

    def __init__(self, image: Image, table_start: int, table_size: int):
        """
        Pass a image to be processed

        :param image: image to be processed
        :type image: Image
        """
        self.image = image
        self.import_stack = self._load_binding_info(table_start, table_size)
        self.actions = self._create_action_list()
        self.lookup_table = {}
        self.link_table = {}
        self.symbol_table = self._load_symbol_table()

    def _load_symbol_table(self) -> List[Symbol]:
        table = []
        for act in self.actions:
            if act.item:
                sym = Symbol.from_values(act.item, act.vmaddr, external=True, ordinal=act.libname)
                table.append(sym)
                self.lookup_table[act.vmaddr] = sym
        return table

    def _create_action_list(self) -> List[action]:
        actions = []
        for bind_command in self.import_stack:
            segment = list(self.image.segments.values())[bind_command.seg_index]
            vm_address = segment.vm_address + bind_command.seg_offset
            try:
                lib = self.image.linked_images[bind_command.lib_ordinal - 1].install_name
            except IndexError:
                lib = str(bind_command.lib_ordinal)
            item = bind_command.name
            actions.append(action(vm_address & 0xFFFFFFFFF, lib, item))
        return actions

    def _load_binding_info(self, table_start: int, table_size: int) -> List[record]:
        read_address = table_start
        import_stack = []
        threaded_stack = []
        uses_threaded_bind = False
        while True:
            if read_address - table_size >= table_start:
                break
            seg_index = 0x0
            seg_offset = 0x0
            lib_ordinal = 0x0
            btype = 0x0
            flags = 0x0
            name = ""
            addend = 0x0
            special_dylib = 0x0
            while True:
                # There are 0xc opcodes total
                # Bitmask opcode byte with 0xF0 to get opcode, 0xF to get value
                binding_opcode = self.image.get_int_at(read_address, 1) & 0xF0
                value = self.image.get_int_at(read_address, 1) & 0x0F
                log.debug_tm(f'{BINDING_OPCODE(binding_opcode).name}: {hex(value)}')
                cmd_start_addr = read_address
                read_address += 1

                if binding_opcode == BINDING_OPCODE.THREADED:
                    if value == BIND_SUBOPCODE_THREADED_SET_BIND_ORDINAL_TABLE_SIZE_ULEB:
                        a_table_size, read_address = self.image.decode_uleb128(read_address)
                        uses_threaded_bind = True
                    elif value == BIND_SUBOPCODE_THREADED_APPLY:
                        pass

                if binding_opcode == BINDING_OPCODE.DONE:
                    import_stack.append(
                        record(cmd_start_addr, seg_index, seg_offset, lib_ordinal, btype, flags, name, addend,
                               special_dylib))
                    break

                elif binding_opcode == BINDING_OPCODE.SET_DYLIB_ORDINAL_IMM:
                    lib_ordinal = value

                elif binding_opcode == BINDING_OPCODE.SET_DYLIB_ORDINAL_ULEB:
                    lib_ordinal, read_address = self.image.decode_uleb128(read_address)

                elif binding_opcode == BINDING_OPCODE.SET_DYLIB_SPECIAL_IMM:
                    special_dylib = 0x1
                    lib_ordinal = value

                elif binding_opcode == BINDING_OPCODE.SET_SYMBOL_TRAILING_FLAGS_IMM:
                    flags = value
                    name = self.image.get_cstr_at(read_address)
                    read_address += len(name) + 1

                elif binding_opcode == BINDING_OPCODE.SET_TYPE_IMM:
                    btype = value

                elif binding_opcode == BINDING_OPCODE.SET_ADDEND_SLEB:
                    addend, read_address = self.image.decode_uleb128(read_address)

                elif binding_opcode == BINDING_OPCODE.SET_SEGMENT_AND_OFFSET_ULEB:
                    seg_index = value
                    seg_offset, read_address = self.image.decode_uleb128(read_address)

                elif binding_opcode == BINDING_OPCODE.ADD_ADDR_ULEB:
                    o, read_address = self.image.decode_uleb128(read_address)
                    seg_offset += o

                elif binding_opcode == BINDING_OPCODE.DO_BIND_ADD_ADDR_ULEB:
                    import_stack.append(
                        record(cmd_start_addr, seg_index, seg_offset, lib_ordinal, btype, flags, name, addend,
                               special_dylib))
                    seg_offset += 8
                    o, read_address = self.image.decode_uleb128(read_address)
                    seg_offset += o

                elif binding_opcode == BINDING_OPCODE.DO_BIND_ADD_ADDR_IMM_SCALED:
                    import_stack.append(
                        record(cmd_start_addr, seg_index, seg_offset, lib_ordinal, btype, flags, name, addend,
                               special_dylib))
                    seg_offset = seg_offset + (value * 8) + 8

                elif binding_opcode == BINDING_OPCODE.DO_BIND_ULEB_TIMES_SKIPPING_ULEB:
                    count, read_address = self.image.decode_uleb128(read_address)
                    skip, read_address = self.image.decode_uleb128(read_address)

                    for i in range(0, count):
                        import_stack.append(
                            record(cmd_start_addr, seg_index, seg_offset, lib_ordinal, btype, flags, name, addend,
                                   special_dylib))
                        seg_offset += skip + 8

                elif binding_opcode == BINDING_OPCODE.DO_BIND:
                    if not uses_threaded_bind:
                        import_stack.append(
                            record(cmd_start_addr, seg_index, seg_offset, lib_ordinal, btype, flags, name, addend,
                                   special_dylib))
                        seg_offset += 8
                    else:
                        threaded_stack.append(
                            record(cmd_start_addr, seg_index, seg_offset, lib_ordinal, btype, flags, name, addend,
                                   special_dylib))
                        seg_offset += 8

        return import_stack



#
#  ktool | kswift
#  demangle.py
#
#
#
#  This file is part of ktool. ktool is free software that
#  is made available under the MIT license. Consult the
#  file "LICENSE" that is distributed together with this file
#  for the exact licensing terms.
#
#  Copyright (c) kat 2022.
#

def demangle(name):
    """
    Very basic, very sloppy bare minimum POC for swift classname demangling

    :param name:
    :return:
    """

    project = ""
    typename = ""
    stage = 0
    skip = False

    for c in name:
        if c.isdigit():
            if skip:
                continue
            else:
                stage += 1
                skip = True
                continue
        else:
            skip = False
            if stage == 0:
                continue
            elif stage == 1:
                project += c
            elif stage == 2:
                typename += c

    return project, typename
